#   Copyright 2017 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography import x509
from cryptography.x509.oid import NameOID
from datetime import datetime
from datetime import timedelta
import fixtures
import mock
import os
import tempfile
import yaml

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from tripleo_common.image import kolla_builder

from tripleoclient import exceptions
from tripleoclient.tests import base
from tripleoclient.v1 import undercloud_config


class TestProcessDriversAndHardwareTypes(base.TestCase):
    def setUp(self):
        super(TestProcessDriversAndHardwareTypes, self).setUp()
        self.conf = mock.Mock(**{key: getattr(undercloud_config.CONF, key)
                                 for key in (
                                     'enabled_hardware_types',
                                     'enable_node_discovery',
                                     'discovery_default_driver',
                                     'ironic_enabled_network_interfaces',
                                     'ironic_default_network_interface')})

    def test_defaults(self):
        env = {}
        undercloud_config._process_drivers_and_hardware_types(self.conf, env)
        self.assertEqual({
            'IronicEnabledNetworkInterfaces': ['flat'],
            'IronicDefaultNetworkInterface': 'flat',
            'IronicEnabledHardwareTypes': ['idrac', 'ilo', 'ipmi', 'redfish'],
            'IronicEnabledBootInterfaces': ['ilo-pxe', 'ipxe', 'pxe'],
            'IronicEnabledBiosInterfaces': ['ilo', 'no-bios', 'redfish'],
            'IronicEnabledDeployInterfaces': ['ansible', 'direct', 'iscsi'],
            'IronicEnabledInspectInterfaces': ['idrac', 'ilo', 'inspector',
                                               'no-inspect', 'redfish'],
            'IronicEnabledManagementInterfaces': ['fake', 'idrac', 'ilo',
                                                  'ipmitool', 'noop',
                                                  'redfish'],
            'IronicEnabledPowerInterfaces': ['fake', 'idrac', 'ilo',
                                             'ipmitool', 'redfish'],
            'IronicEnabledRaidInterfaces': ['idrac', 'no-raid'],
            'IronicEnabledVendorInterfaces': ['idrac', 'ipmitool', 'no-vendor']
        }, env)

    def test_one_hardware_type_with_discovery(self):
        env = {}
        self.conf.enabled_hardware_types = ['redfish']
        self.conf.enable_node_discovery = True

        undercloud_config._process_drivers_and_hardware_types(self.conf, env)
        self.assertEqual({
            'IronicEnabledNetworkInterfaces': ['flat'],
            'IronicDefaultNetworkInterface': 'flat',
            # ipmi added because it's the default discovery driver
            'IronicEnabledHardwareTypes': ['ipmi', 'redfish'],
            'IronicEnabledBootInterfaces': ['ipxe', 'pxe'],
            'IronicEnabledBiosInterfaces': ['no-bios', 'redfish'],
            'IronicEnabledDeployInterfaces': ['ansible', 'direct', 'iscsi'],
            'IronicEnabledInspectInterfaces': ['inspector', 'no-inspect',
                                               'redfish'],
            'IronicEnabledManagementInterfaces': ['fake', 'ipmitool',
                                                  'noop', 'redfish'],
            'IronicEnabledPowerInterfaces': ['fake', 'ipmitool', 'redfish'],
            'IronicEnabledRaidInterfaces': ['no-raid'],
            'IronicEnabledVendorInterfaces': ['ipmitool', 'no-vendor'],
            'IronicInspectorDiscoveryDefaultDriver': 'ipmi',
            'IronicInspectorEnableNodeDiscovery': True
        }, env)

    def test_all_hardware_types(self):
        env = {}
        self.conf.enabled_hardware_types = (
            self.conf.enabled_hardware_types + ['staging-ovirt', 'snmp',
                                                'irmc', 'xclarity',
                                                'fake-hardware']
        )

        undercloud_config._process_drivers_and_hardware_types(self.conf, env)
        self.assertEqual({
            'IronicEnabledNetworkInterfaces': ['flat'],
            'IronicDefaultNetworkInterface': 'flat',
            'IronicEnabledHardwareTypes': ['fake-hardware', 'idrac', 'ilo',
                                           'ipmi', 'irmc', 'redfish', 'snmp',
                                           'staging-ovirt', 'xclarity'],
            'IronicEnabledBootInterfaces': ['fake', 'ilo-pxe', 'ipxe',
                                            'irmc-pxe', 'pxe'],
            'IronicEnabledBiosInterfaces': ['ilo', 'irmc',
                                            'no-bios', 'redfish'],
            'IronicEnabledDeployInterfaces': ['ansible', 'direct', 'fake',
                                              'iscsi'],
            'IronicEnabledInspectInterfaces': ['idrac', 'ilo', 'inspector',
                                               'irmc', 'no-inspect',
                                               'redfish'],
            'IronicEnabledManagementInterfaces': ['fake', 'idrac',
                                                  'ilo', 'ipmitool', 'irmc',
                                                  'noop', 'redfish',
                                                  'staging-ovirt', 'xclarity'],
            'IronicEnabledPowerInterfaces': ['fake', 'idrac',
                                             'ilo', 'ipmitool', 'irmc',
                                             'redfish', 'snmp',
                                             'staging-ovirt', 'xclarity'],
            'IronicEnabledRaidInterfaces': ['idrac', 'no-raid'],
            'IronicEnabledVendorInterfaces': ['idrac', 'ipmitool', 'no-vendor']
        }, env)


class TestBaseNetworkSettings(base.TestCase):
    def setUp(self):
        super(TestBaseNetworkSettings, self).setUp()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        # don't actually load config from ~/undercloud.conf
        self.mock_config_load = self.useFixture(
            fixtures.MockPatch('tripleoclient.utils.load_config'))
        self.conf.config(local_ip='192.168.24.1/24',
                         undercloud_admin_host='192.168.24.3',
                         undercloud_public_host='192.168.24.2',
                         undercloud_nameservers=['10.10.10.10', '10.10.10.11'])
        # ctlplane network - config group options
        self.grp0 = cfg.OptGroup(name='ctlplane-subnet',
                                 title='ctlplane-subnet')
        self.opts = [cfg.StrOpt('cidr'),
                     cfg.ListOpt('dhcp_start'),
                     cfg.ListOpt('dhcp_end'),
                     cfg.ListOpt('dhcp_exclude'),
                     cfg.StrOpt('inspection_iprange'),
                     cfg.StrOpt('gateway'),
                     cfg.BoolOpt('masquerade'),
                     cfg.ListOpt('host_routes',
                                 item_type=cfg.types.Dict(bounds=True),
                                 bounds=True,),
                     cfg.ListOpt('dns_nameservers')]
        self.conf.register_opts(self.opts, group=self.grp0)
        self.grp1 = cfg.OptGroup(name='subnet1', title='subnet1')
        self.grp2 = cfg.OptGroup(name='subnet2', title='subnet2')
        self.conf.config(cidr='192.168.24.0/24',
                         dhcp_start='192.168.24.5',
                         dhcp_end='192.168.24.24',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.24.100,192.168.24.120',
                         gateway='192.168.24.1',
                         masquerade=False,
                         host_routes=[],
                         dns_nameservers=[],
                         group='ctlplane-subnet')


class TestNetworkSettings(TestBaseNetworkSettings):
    def test_default(self):
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.5', 'end': '192.168.24.24'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_ipv6_control_plane_stateless_default(self):
        env = {}
        self.conf.config(local_ip='fd12:3456:789a:1::2/64',
                         undercloud_admin_host='fd12:3456:789a:1::3',
                         undercloud_public_host='fd12:3456:789a:1::4')
        self.conf.config(cidr='fd12:3456:789a:1::/64',
                         dhcp_start='fd12:3456:789a:1::10',
                         dhcp_end='fd12:3456:789a:1::20',
                         dhcp_exclude=[],
                         dns_nameservers=['fd12:3456:789a:1::5',
                                          'fd12:3456:789a:1::6'],
                         inspection_iprange=('fd12:3456:789a:1::30,'
                                             'fd12:3456:789a:1::40'),
                         gateway='fd12:3456:789a:1::1',
                         masquerade=False,
                         host_routes=[],
                         group='ctlplane-subnet')
        undercloud_config._process_network_args(env)
        expected = {
            'NovaIPv6': True,
            'RabbitIPv6': True,
            'MemcachedIPv6': True,
            'RedisIPv6': True,
            'MysqlIPv6': True,
            'IronicIpVersion': '6',
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': 'fd12:3456:789a:1::1',
                 'host_routes': [],
                 'ip_range': 'fd12:3456:789a:1::,static',
                 'netmask': 'ffff:ffff:ffff:ffff::',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'fd12:3456:789a:1::/64': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': 'fd12:3456:789a:1::10',
                         'end': 'fd12:3456:789a:1::20'}],
                    'DnsNameServers': ['fd12:3456:789a:1::5',
                                       'fd12:3456:789a:1::6'],
                    'HostRoutes': [],
                    'NetworkCidr': 'fd12:3456:789a:1::/64',
                    'NetworkGateway': 'fd12:3456:789a:1::1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_ipv6_control_plane_stateful(self):
        env = {}
        self.conf.config(local_ip='fd12:3456:789a:1::2/64',
                         undercloud_admin_host='fd12:3456:789a:1::3',
                         undercloud_public_host='fd12:3456:789a:1::4',
                         ipv6_address_mode='dhcpv6-stateful')
        self.conf.config(cidr='fd12:3456:789a:1::/64',
                         dhcp_start='fd12:3456:789a:1::10',
                         dhcp_end='fd12:3456:789a:1::20',
                         dhcp_exclude=[],
                         dns_nameservers=['fd12:3456:789a:1::5',
                                          'fd12:3456:789a:1::6'],
                         inspection_iprange=('fd12:3456:789a:1::30,'
                                             'fd12:3456:789a:1::40'),
                         gateway='fd12:3456:789a:1::1',
                         masquerade=False,
                         host_routes=[],
                         group='ctlplane-subnet')
        undercloud_config._process_network_args(env)
        expected = {
            'NovaIPv6': True,
            'RabbitIPv6': True,
            'MemcachedIPv6': True,
            'RedisIPv6': True,
            'MysqlIPv6': True,
            'IronicIpVersion': '6',
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': 'fd12:3456:789a:1::1',
                 'host_routes': [],
                 'ip_range': 'fd12:3456:789a:1::30,fd12:3456:789a:1::40',
                 'netmask': 'ffff:ffff:ffff:ffff::',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'fd12:3456:789a:1::/64': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': 'fd12:3456:789a:1::10',
                         'end': 'fd12:3456:789a:1::20'}],
                    'DnsNameServers': ['fd12:3456:789a:1::5',
                                       'fd12:3456:789a:1::6'],
                    'HostRoutes': [],
                    'NetworkCidr': 'fd12:3456:789a:1::/64',
                    'NetworkGateway': 'fd12:3456:789a:1::1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateful',
        }
        self.assertEqual(expected, env)

    def test_nameserver_toomany_fail(self):
        env = {}
        self.conf.config(undercloud_nameservers=['1.1.1.1', '1.1.1.2',
                                                 '1.1.1.3', '1.1.1.4',
                                                 '1.1.1.5', '1.1.1.6'])
        self.assertRaises(exceptions.InvalidConfiguration,
                          undercloud_config._process_network_args,
                          env)

    def test_undercloud_ips_duplicated_fail(self):
        env = {}

        # local_ip == undercloud_admin_host
        self.conf.config(local_ip='192.168.24.1/24',
                         undercloud_admin_host='192.168.24.1',
                         undercloud_public_host='192.168.24.2',
                         generate_service_certificate=True)
        self.assertRaises(exceptions.InvalidConfiguration,
                          undercloud_config._process_network_args,
                          env)

        # local_ip == undercloud_public_host
        self.conf.config(local_ip='192.168.24.1/24',
                         undercloud_admin_host='192.168.24.3',
                         undercloud_public_host='192.168.24.1',
                         generate_service_certificate=True)
        undercloud_config._process_network_args(env)

        # undercloud_admin_host == undercloud_public_host
        self.conf.config(local_ip='192.168.24.1/24',
                         undercloud_admin_host='192.168.24.2',
                         undercloud_public_host='192.168.24.2',
                         generate_service_certificate=True)
        undercloud_config._process_network_args(env)

        # We do not care about ip duplication when ssl is disabled
        self.conf.config(local_ip='192.168.24.1/24',
                         undercloud_admin_host='192.168.24.1',
                         undercloud_public_host='192.168.24.2',
                         generate_service_certificate=False,
                         undercloud_service_certificate='')
        undercloud_config._process_network_args(env)

    def test_start_end_all_addresses(self):
        self.conf.config(dhcp_start='192.168.24.0',
                         dhcp_end='192.168.24.255',
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.4', 'end': '192.168.24.99'},
                        {'start': '192.168.24.121', 'end': '192.168.24.254'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_ignore_dhcp_start_end_if_default_but_cidr_not_default(self):
        self.conf.config(cidr='192.168.10.0/24',
                         inspection_iprange='192.168.10.100,192.168.10.120',
                         gateway='192.168.10.1',
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.10.1',
                 'host_routes': [],
                 'ip_range': '192.168.10.100,192.168.10.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.10.2', 'end': '192.168.10.99'},
                        {'start': '192.168.10.121', 'end': '192.168.10.254'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_dhcp_exclude(self):
        self.conf.config(cidr='192.168.10.0/24',
                         inspection_iprange='192.168.10.100,192.168.10.120',
                         gateway='192.168.10.1',
                         dhcp_exclude=['192.168.10.50',
                                       '192.168.10.80-192.168.10.89'],
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.10.1',
                 'host_routes': [],
                 'ip_range': '192.168.10.100,192.168.10.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.10.2', 'end': '192.168.10.49'},
                        {'start': '192.168.10.51', 'end': '192.168.10.79'},
                        {'start': '192.168.10.90', 'end': '192.168.10.99'},
                        {'start': '192.168.10.121', 'end': '192.168.10.254'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_no_dhcp_start_no_dhcp_end(self):
        self.conf.config(dhcp_start=[],
                         dhcp_end=[],
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.4', 'end': '192.168.24.99'},
                        {'start': '192.168.24.121', 'end': '192.168.24.254'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'}},
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_dhcp_start_no_dhcp_end(self):
        self.conf.config(dhcp_start='192.168.24.10',
                         dhcp_end=[],
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.10', 'end': '192.168.24.99'},
                        {'start': '192.168.24.121', 'end': '192.168.24.254'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_dhcp_end_no_dhcp_start(self):
        self.conf.config(dhcp_start=[],
                         dhcp_end='192.168.24.220',
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500}],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.4', 'end': '192.168.24.99'},
                        {'start': '192.168.24.121', 'end': '192.168.24.220'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_routed_network(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.register_opts(self.opts, group=self.grp2)
        self.conf.config(masquerade=True,
                         dns_nameservers=['10.1.1.100', '10.1.1.101'],
                         group='ctlplane-subnet')
        self.conf.config(cidr='192.168.10.0/24',
                         dhcp_start='192.168.10.10',
                         dhcp_end='192.168.10.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.10.100,192.168.10.189',
                         gateway='192.168.10.254',
                         dns_nameservers=['10.2.2.100', '10.2.2.101'],
                         host_routes=[],
                         masquerade=True,
                         group='subnet1')
        self.conf.config(cidr='192.168.20.0/24',
                         dhcp_start='192.168.20.10',
                         dhcp_end='192.168.20.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.20.100,192.168.20.189',
                         gateway='192.168.20.254',
                         dns_nameservers=['10.3.3.100', '10.3.3.101'],
                         host_routes=[],
                         masquerade=True,
                         group='subnet2')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [
                {'destination': '192.168.10.0/24', 'nexthop': '192.168.24.1'},
                {'destination': '192.168.20.0/24', 'nexthop': '192.168.24.1'}],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500},
                {'gateway': '192.168.10.254',
                 'host_routes': [],
                 'ip_range': '192.168.10.100,192.168.10.189',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet1',
                 'mtu': 1500},
                {'gateway': '192.168.20.254',
                 'host_routes': [],
                 'ip_range': '192.168.20.100,192.168.20.189',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet2',
                 'mtu': 1500}
            ],
            'MasqueradeNetworks': {
                '192.168.10.0/24': ['192.168.24.0/24',
                                    '192.168.10.0/24',
                                    '192.168.20.0/24'],
                '192.168.20.0/24': ['192.168.24.0/24',
                                    '192.168.10.0/24',
                                    '192.168.20.0/24'],
                '192.168.24.0/24': ['192.168.24.0/24',
                                    '192.168.10.0/24',
                                    '192.168.20.0/24']},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'subnet1',
                                   '192.168.20.0/24': 'subnet2',
                                   '192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                # The ctlplane-subnet subnet have defaults
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.5', 'end': '192.168.24.24'}],
                    'DnsNameServers': ['10.1.1.100', '10.1.1.101'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
                'subnet1': {
                    'AllocationPools': [
                        {'start': '192.168.10.10', 'end': '192.168.10.99'}],
                    'DnsNameServers': ['10.2.2.100', '10.2.2.101'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.254'},
                'subnet2': {
                    'AllocationPools': [
                        {'start': '192.168.20.10', 'end': '192.168.20.99'}],
                    'DnsNameServers': ['10.3.3.100', '10.3.3.101'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.20.0/24',
                    'NetworkGateway': '192.168.20.254'}
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_routed_network_no_masquerading(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.register_opts(self.opts, group=self.grp2)
        self.conf.config(cidr='192.168.10.0/24',
                         dhcp_start='192.168.10.10',
                         dhcp_end='192.168.10.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.10.100,192.168.10.189',
                         gateway='192.168.10.254',
                         dns_nameservers=[],
                         host_routes=[],
                         group='subnet1')
        self.conf.config(cidr='192.168.20.0/24',
                         dhcp_start='192.168.20.10',
                         dhcp_end='192.168.20.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.20.100,192.168.20.189',
                         gateway='192.168.20.254',
                         dns_nameservers=[],
                         host_routes=[],
                         group='subnet2')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [
                {'destination': '192.168.10.0/24', 'nexthop': '192.168.24.1'},
                {'destination': '192.168.20.0/24', 'nexthop': '192.168.24.1'}],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500},
                {'gateway': '192.168.10.254',
                 'host_routes': [],
                 'ip_range': '192.168.10.100,192.168.10.189',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet1',
                 'mtu': 1500},
                {'gateway': '192.168.20.254',
                 'host_routes': [],
                 'ip_range': '192.168.20.100,192.168.20.189',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet2',
                 'mtu': 1500}
            ],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'subnet1',
                                   '192.168.20.0/24': 'subnet2',
                                   '192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                # The ctlplane-subnet subnet have defaults
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.5', 'end': '192.168.24.24'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
                'subnet1': {
                    'AllocationPools': [
                        {'start': '192.168.10.10', 'end': '192.168.10.99'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.254'},
                'subnet2': {
                    'AllocationPools': [
                        {'start': '192.168.20.10', 'end': '192.168.20.99'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.20.0/24',
                    'NetworkGateway': '192.168.20.254'}
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_no_allocation_pool_on_remote_network(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1'])
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.config(cidr='192.168.10.0/24',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.10.200,192.168.10.254',
                         gateway='192.168.10.254',
                         dns_nameservers=[],
                         host_routes=[],
                         masquerade=False,
                         group='subnet1')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [
                {'destination': '192.168.10.0/24', 'nexthop': '192.168.24.1'}],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500},
                {'gateway': '192.168.10.254',
                 'host_routes': [],
                 'ip_range': '192.168.10.200,192.168.10.254',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet1',
                 'mtu': 1500},
            ],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'subnet1',
                                   '192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                # The ctlplane-subnet subnet have defaults
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.5', 'end': '192.168.24.24'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
                'subnet1': {
                    'AllocationPools': [
                        {'start': '192.168.10.1', 'end': '192.168.10.199'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.254'}
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_no_allocation_pool_on_remote_network_three_pools(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1'])
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.config(cidr='192.168.10.0/24',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.10.100,192.168.10.199',
                         gateway='192.168.10.222',
                         dns_nameservers=[],
                         host_routes=[],
                         masquerade=False,
                         group='subnet1')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [
                {'destination': '192.168.10.0/24', 'nexthop': '192.168.24.1'}],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500},
                {'gateway': '192.168.10.222',
                 'host_routes': [],
                 'ip_range': '192.168.10.100,192.168.10.199',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet1',
                 'mtu': 1500},
            ],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'subnet1',
                                   '192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                # The ctlplane-subnet subnet have defaults
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.5', 'end': '192.168.24.24'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
                'subnet1': {
                    'AllocationPools': [
                        {'start': '192.168.10.1', 'end': '192.168.10.99'},
                        {'start': '192.168.10.200', 'end': '192.168.10.221'},
                        {'start': '192.168.10.223', 'end': '192.168.10.254'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.222'}
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless',
        }
        self.assertEqual(expected, env)

    def test_additional_host_routes(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
        self.conf.config(host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.24.1'}],
                         group='ctlplane-subnet')
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.register_opts(self.opts, group=self.grp2)
        self.conf.config(cidr='192.168.10.0/24',
                         dhcp_start='192.168.10.10',
                         dhcp_end='192.168.10.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.10.100,192.168.10.189',
                         gateway='192.168.10.254',
                         dns_nameservers=[],
                         host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.10.254'}],
                         group='subnet1')
        self.conf.config(cidr='192.168.20.0/24',
                         dhcp_start='192.168.20.10',
                         dhcp_end='192.168.20.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.20.100,192.168.20.189',
                         gateway='192.168.20.254',
                         dns_nameservers=[],
                         host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.20.254'}],
                         group='subnet2')
        env = {}
        undercloud_config._process_network_args(env)
        expected = {
            'ControlPlaneStaticRoutes': [
                {'destination': '192.168.10.0/24', 'nexthop': '192.168.24.1'},
                {'destination': '192.168.20.0/24', 'nexthop': '192.168.24.1'},
                {'destination': '10.10.10.254/32', 'nexthop': '192.168.24.1'}],
            'IronicInspectorSubnets': [
                {'gateway': '192.168.24.1',
                 'host_routes': [{'destination': '10.10.10.254/32',
                                  'nexthop': '192.168.24.1'}],
                 'ip_range': '192.168.24.100,192.168.24.120',
                 'netmask': '255.255.255.0',
                 'tag': 'ctlplane-subnet',
                 'mtu': 1500},
                {'gateway': '192.168.10.254',
                 'host_routes': [{'destination': '10.10.10.254/32',
                                  'nexthop': '192.168.10.254'}],
                 'ip_range': '192.168.10.100,192.168.10.189',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet1',
                 'mtu': 1500},
                {'gateway': '192.168.20.254',
                 'host_routes': [{'destination': '10.10.10.254/32',
                                  'nexthop': '192.168.20.254'}],
                 'ip_range': '192.168.20.100,192.168.20.189',
                 'netmask': '255.255.255.0',
                 'tag': 'subnet2',
                 'mtu': 1500}
            ],
            'MasqueradeNetworks': {},
            'PortPhysnetCidrMap': {'192.168.10.0/24': 'subnet1',
                                   '192.168.20.0/24': 'subnet2',
                                   '192.168.24.0/24': 'ctlplane'},
            'UndercloudCtlplaneSubnets': {
                # The ctlplane-subnet subnet have defaults
                'ctlplane-subnet': {
                    'AllocationPools': [
                        {'start': '192.168.24.5', 'end': '192.168.24.24'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [{'destination': '10.10.10.254/32',
                                    'nexthop': '192.168.24.1'}],
                    'NetworkCidr': '192.168.24.0/24',
                    'NetworkGateway': '192.168.24.1'},
                'subnet1': {
                    'AllocationPools': [
                        {'start': '192.168.10.10', 'end': '192.168.10.99'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [{'destination': '10.10.10.254/32',
                                    'nexthop': '192.168.10.254'}],
                    'NetworkCidr': '192.168.10.0/24',
                    'NetworkGateway': '192.168.10.254'},
                'subnet2': {
                    'AllocationPools': [
                        {'start': '192.168.20.10', 'end': '192.168.20.99'}],
                    'DnsNameServers': ['10.10.10.10', '10.10.10.11'],
                    'HostRoutes': [{'destination': '10.10.10.254/32',
                                    'nexthop': '192.168.20.254'}],
                    'NetworkCidr': '192.168.20.0/24',
                    'NetworkGateway': '192.168.20.254'}
            },
            'UndercloudCtlplaneIPv6AddressMode': 'dhcpv6-stateless'
        }
        self.assertEqual(expected, env)

    def test_generate_inspection_subnets(self):
        result = undercloud_config._generate_inspection_subnets()
        expected = [{'gateway': '192.168.24.1',
                     'host_routes': [],
                     'ip_range': '192.168.24.100,192.168.24.120',
                     'mtu': 1500,
                     'netmask': '255.255.255.0',
                     'tag': 'ctlplane-subnet'}]
        self.assertEqual(expected, result)

    def test_generate_inspection_subnets_invalid(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1'])
        self.conf.config(host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.24.1'}],
                         group='ctlplane-subnet')
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.config(group='subnet1')
        self.assertRaises(exceptions.DeploymentError,
                          undercloud_config._generate_inspection_subnets)

    def test__env_set_undercloud_ctlplane_networks_attribues(self):
        self.conf.config(local_subnet='ctlplane-subnet',
                         local_mtu=1444,
                         undercloud_nameservers=['192.168.24.253',
                                                 '192.168.24.252'])
        self.conf.config(cidr='192.168.24.0/24',
                         gateway='192.168.24.254',
                         host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.24.1'}],
                         group='ctlplane-subnet')
        env = {}
        undercloud_config._env_set_undercloud_ctlplane_networks_attribues(env)
        expected = {
            'CtlplaneNetworkAttributes': {
                'network': {'mtu': 1444},
                'subnets': {
                    'ctlplane-subnet': {
                        'cidr': '192.168.24.0/24',
                        'dns_nameservers': ['192.168.24.253',
                                            '192.168.24.252'],
                        'gateway_ip': '192.168.24.254',
                        'host_routes': [{'destination': '10.10.10.254/32',
                                         'nexthop': '192.168.24.1'}],
                        'tags': []}}}}
        self.assertEqual(expected, env)

    def test__env_set_undercloud_ctlplane_networks_attribues_routed(self):
        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
        self.conf.config(host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.24.1'}],
                         group='ctlplane-subnet')
        self.conf.register_opts(self.opts, group=self.grp1)
        self.conf.register_opts(self.opts, group=self.grp2)
        self.conf.config(cidr='192.168.10.0/24',
                         dhcp_start='192.168.10.10',
                         dhcp_end='192.168.10.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.10.100,192.168.10.189',
                         gateway='192.168.10.254',
                         dns_nameservers=[],
                         host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.10.254'}],
                         group='subnet1')
        self.conf.config(cidr='192.168.20.0/24',
                         dhcp_start='192.168.20.10',
                         dhcp_end='192.168.20.99',
                         dhcp_exclude=[],
                         inspection_iprange='192.168.20.100,192.168.20.189',
                         gateway='192.168.20.254',
                         dns_nameservers=[],
                         host_routes=[{'destination': '10.10.10.254/32',
                                       'nexthop': '192.168.20.254'}],
                         group='subnet2')
        env = {}
        undercloud_config._env_set_undercloud_ctlplane_networks_attribues(env)
        expected = {
            'CtlplaneNetworkAttributes': {
                'network': {'mtu': 1500},
                'subnets': {
                    'ctlplane-subnet': {
                        'cidr': '192.168.24.0/24',
                        'dns_nameservers': ['10.10.10.10', '10.10.10.11'],
                        'gateway_ip': '192.168.24.1',
                        'host_routes': [{'destination': '192.168.10.0/24',
                                         'nexthop': '192.168.24.1'},
                                        {'destination': '192.168.20.0/24',
                                         'nexthop': '192.168.24.1'},
                                        {'destination': '10.10.10.254/32',
                                         'nexthop': '192.168.24.1'}],
                        'tags': []}}}}
        self.assertEqual(expected, env)


class TestChronySettings(TestBaseNetworkSettings):
    def test_default(self):
        env = {}
        undercloud_config._process_chrony_acls(env)
        expected = {
            'ChronyAclRules': ['allow 192.168.24.0/24'],
        }
        self.assertEqual(expected, env)


class TestTLSSettings(base.TestCase):
    def test_public_host_with_ip_should_give_ip_endpoint_environment(self):
        expected_env_file = os.path.join(
            undercloud_config.THT_HOME,
            "environments/ssl/tls-endpoints-public-ip.yaml")

        resulting_env_file1 = undercloud_config._get_tls_endpoint_environment(
            '127.0.0.1', undercloud_config.THT_HOME)

        self.assertEqual(expected_env_file, resulting_env_file1)

        resulting_env_file2 = undercloud_config._get_tls_endpoint_environment(
            '192.168.1.1', undercloud_config.THT_HOME)

        self.assertEqual(expected_env_file, resulting_env_file2)

    def test_public_host_with_fqdn_should_give_dns_endpoint_environment(self):
        expected_env_file = os.path.join(
            undercloud_config.THT_HOME,
            "environments/ssl/tls-endpoints-public-dns.yaml")

        resulting_env_file1 = undercloud_config._get_tls_endpoint_environment(
            'controller-1', undercloud_config.THT_HOME)

        self.assertEqual(expected_env_file, resulting_env_file1)

        resulting_env_file2 = undercloud_config._get_tls_endpoint_environment(
            'controller-1.tripleodomain.com', undercloud_config.THT_HOME)

        self.assertEqual(expected_env_file, resulting_env_file2)

    def get_certificate_and_private_key(self):
        private_key = rsa.generate_private_key(public_exponent=3,
                                               key_size=1024,
                                               backend=default_backend())
        issuer = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, u"FI"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, u"Helsinki"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"Some Company"),
            x509.NameAttribute(NameOID.COMMON_NAME, u"Test Certificate"),
        ])
        cert_builder = x509.CertificateBuilder(
            issuer_name=issuer, subject_name=issuer,
            public_key=private_key.public_key(),
            serial_number=x509.random_serial_number(),
            not_valid_before=datetime.utcnow(),
            not_valid_after=datetime.utcnow() + timedelta(days=10)
        )
        cert = cert_builder.sign(private_key,
                                 hashes.SHA256(),
                                 default_backend())
        cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
        key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption())
        return cert_pem, key_pem

    def test_get_dict_with_cert_and_key_from_bundled_pem(self):
        cert_pem, key_pem = self.get_certificate_and_private_key()

        with tempfile.NamedTemporaryFile() as tempbundle:
            tempbundle.write(cert_pem)
            tempbundle.write(key_pem)
            tempbundle.seek(0)

            tls_parameters = undercloud_config._get_public_tls_parameters(
                tempbundle.name)

        self.assertEqual(cert_pem, tls_parameters['SSLCertificate'])
        self.assertEqual(key_pem, tls_parameters['SSLKey'])

    def test_get_tls_parameters_fails_cause_of_missing_cert(self):
        _, key_pem = self.get_certificate_and_private_key()

        with tempfile.NamedTemporaryFile() as tempbundle:
            tempbundle.write(key_pem)
            tempbundle.seek(0)

            self.assertRaises(ValueError,
                              undercloud_config._get_public_tls_parameters,
                              tempbundle.name)

    def test_get_tls_parameters_fails_cause_of_missing_key(self):
        cert_pem, _ = self.get_certificate_and_private_key()

        with tempfile.NamedTemporaryFile() as tempbundle:
            tempbundle.write(cert_pem)
            tempbundle.seek(0)

            self.assertRaises(ValueError,
                              undercloud_config._get_public_tls_parameters,
                              tempbundle.name)

    def test_get_tls_parameters_fails_cause_of_unexistent_file(self):
        self.assertRaises(IOError,
                          undercloud_config._get_public_tls_parameters,
                          '/tmp/unexistent-file-12345.pem')


class TestContainerImageConfig(base.TestCase):
    def setUp(self):
        super(TestContainerImageConfig, self).setUp()
        conf_keys = (
            'container_images_file',
        )
        self.conf = mock.Mock(**{key: getattr(undercloud_config.CONF, key)
                                 for key in conf_keys})

    @mock.patch('shutil.copy')
    def test_defaults(self, mock_copy):
        env = {}
        deploy_args = []
        cip_default = getattr(kolla_builder,
                              'CONTAINER_IMAGE_PREPARE_PARAM', None)
        self.addCleanup(setattr, kolla_builder,
                        'CONTAINER_IMAGE_PREPARE_PARAM', cip_default)

        setattr(kolla_builder, 'CONTAINER_IMAGE_PREPARE_PARAM', [{
            'set': {
                'namespace': 'one',
                'name_prefix': 'two',
                'name_suffix': 'three',
                'tag': 'four',
            },
            'tag_from_label': 'five',
        }])

        undercloud_config._container_images_config(self.conf, deploy_args,
                                                   env, None)
        self.assertEqual([], deploy_args)
        cip = env['ContainerImagePrepare'][0]
        set = cip['set']

        self.assertEqual(
            'one', set['namespace'])
        self.assertEqual(
            'two', set['name_prefix'])
        self.assertEqual(
            'three', set['name_suffix'])
        self.assertEqual(
            'four', set['tag'])
        self.assertEqual(
            'five', cip['tag_from_label'])

    @mock.patch('shutil.copy')
    def test_container_images_file(self, mock_copy):
        env = {}
        deploy_args = []
        self.conf.container_images_file = '/tmp/container_images_file.yaml'
        undercloud_config._container_images_config(self.conf, deploy_args,
                                                   env, None)
        self.assertEqual(['-e', '/tmp/container_images_file.yaml'],
                         deploy_args)
        self.assertEqual({}, env)

    @mock.patch('shutil.copy')
    def test_custom(self, mock_copy):
        env = {}
        deploy_args = []
        with tempfile.NamedTemporaryFile(mode='w') as f:
            yaml.dump({
                'parameter_defaults': {'ContainerImagePrepare': [{
                    'set': {
                        'namespace': 'one',
                        'name_prefix': 'two',
                        'name_suffix': 'three',
                        'tag': 'four',
                    },
                    'tag_from_label': 'five',
                }]}
            }, f)
            self.conf.container_images_file = f.name
            cif_name = f.name

            undercloud_config._container_images_config(
                self.conf, deploy_args, env, None)
        self.assertEqual(['-e', cif_name], deploy_args)



#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import mock

import ironic_inspector_client
from osc_lib.tests import utils


class FakeInspectorClient(object):
    def __init__(self, states=None, data=None):
        self.states = states or {}
        self.data = data or {}
        self.on_introspection = []

    def introspect(self, uuid):
        self.on_introspection.append(uuid)

    def get_status(self, uuid):
        try:
            return self.states[uuid]
        except KeyError:
            raise ironic_inspector_client.ClientError(mock.Mock())

    def get_data(self, uuid):
        try:
            return self.data[uuid]
        except KeyError:
            raise ironic_inspector_client.ClientError(mock.Mock())

    def wait_for_finish(self, uuids):
        return {uuid: self.states[uuid] for uuid in uuids}


class ClientWrapper(object):

    def __init__(self):
        self._instance = None


class TestBaremetal(utils.TestCommand):

    def setUp(self):
        super(TestBaremetal, self).setUp()

        self.app.client_manager.auth_ref = mock.Mock(auth_token="TOKEN")
        self.app.client_manager.baremetal = mock.Mock()
        self.app.client_manager.image = mock.Mock()
        self.app.client_manager.baremetal_introspection = FakeInspectorClient()
        self.app.client_manager._region_name = "Arcadia"
        self.app.client_manager.session = mock.Mock()
        self.app.client_manager.workflow_engine = mock.Mock()
        self.app.client_manager.tripleoclient = ClientWrapper()

    def tearDown(self):
        super(TestBaremetal, self).tearDown()

        mock.patch.stopall()



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import logging
import simplejson
import yaml

from osc_lib.i18n import _

from tripleoclient import command
from tripleoclient import constants
from tripleoclient import exceptions
from tripleoclient import utils
from tripleoclient.workflows import parameters


class SetParameters(command.Command):
    """Set a parameters for a plan"""

    log = logging.getLogger(__name__ + ".CreatePlan")

    def get_parser(self, prog_name):
        parser = super(SetParameters, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            help=_('The name of the plan, which is used for the Swift '
                   'container, Mistral environment and Heat stack names.'))
        parser.add_argument('file_in', type=argparse.FileType('r'))

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        if parsed_args.file_in.name.endswith('.json'):
            params = simplejson.load(parsed_args.file_in)
        elif parsed_args.file_in.name.endswith('.yaml'):
            params = yaml.safe_load(parsed_args.file_in)
        else:
            raise exceptions.InvalidConfiguration(
                _("Invalid file extension for %s, must be json or yaml") %
                parsed_args.file_in.name)

        if 'parameter_defaults' in params:
            params = params['parameter_defaults']

        with utils.TempDirs() as tmp:
            utils.run_ansible_playbook(
                playbook='cli-update-params.yaml',
                inventory='localhost,',
                workdir=tmp,
                playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                verbosity=utils.playbook_verbosity(self=self),
                extra_vars={
                    "container": parsed_args.name
                },
                extra_vars_file={
                    "parameters": params
                }
            )


class GenerateFencingParameters(command.Command):
    """Generate fencing parameters"""

    log = logging.getLogger(__name__ + ".GenerateFencing")

    def get_parser(self, prog_name):
        parser = super(GenerateFencingParameters, self).get_parser(prog_name)
        parser.add_argument('-a', '--action', dest='fence_action',
                            help=_('DEPRECATED: This option is ignored.'))
        parser.add_argument('--delay', type=int,
                            help=_('Wait DELAY seconds before fencing is '
                                   'started'))
        parser.add_argument('--ipmi-lanplus',
                            dest='ipmi_lanplus',
                            default=True,
                            action='store_true',
                            help=_('DEPRECATED: This is the default.'))
        parser.add_argument('--ipmi-no-lanplus',
                            dest='ipmi_lanplus',
                            action='store_false',
                            help=_('Do not use Lanplus. Defaults to: false'))
        parser.add_argument('--ipmi-cipher', type=int,
                            help=_('Ciphersuite to use (same as ipmitool -C '
                                   'parameter.'))
        parser.add_argument('--ipmi-level',
                            help=_('Privilegel level on IPMI device. Valid '
                                   'levels: callback, user, operator, '
                                   'administrator.'))
        parser.add_argument('--output', type=argparse.FileType('w'),
                            help=_('Write parameters to a file'))
        parser.add_argument('instackenv', type=argparse.FileType('r'))
        return parser

    def take_action(self, parsed_args):
        nodes_config = utils.parse_env_file(parsed_args.instackenv)
        parsed_args.instackenv.close()
        result = parameters.generate_fencing_parameters(
            clients=self.app.client_manager,
            nodes_json=nodes_config,
            delay=parsed_args.delay,
            ipmi_level=parsed_args.ipmi_level,
            ipmi_cipher=parsed_args.ipmi_cipher,
            ipmi_lanplus=parsed_args.ipmi_lanplus,
        )

        fencing_parameters = yaml.safe_dump(result, default_flow_style=False)
        if parsed_args.output:
            parsed_args.output.write(fencing_parameters)
            parsed_args.output.close()
        else:
            print(fencing_parameters)



#   Copyright 2017 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import collections
import os
import sys

from tripleo_common.exception import NotFound
from tripleo_common.utils import roles as rolesutils

from tripleoclient import command
from tripleoclient.constants import TRIPLEO_HEAT_TEMPLATES


class RolesBaseCommand(command.Command):
    auth_required = False

    def get_parser(self, prog_name):
        parser = super(RolesBaseCommand, self).get_parser(prog_name)
        path = os.path.join(TRIPLEO_HEAT_TEMPLATES, 'roles')
        parser.add_argument('--roles-path', metavar='<roles directory>',
                            default=path,
                            help='Filesystem path containing the role yaml  '
                                 'files. By default this is {}'.format(path))
        return parser


class RolesGenerate(RolesBaseCommand):
    """Generate roles_data.yaml file"""
    def get_parser(self, prog_name):
        parser = super(RolesGenerate, self).get_parser(prog_name)
        parser.add_argument('-o', '--output-file', metavar='<output file>',
                            help='File to capture all output to. For example, '
                                 'roles_data.yaml')
        parser.add_argument('--skip-validate', action='store_false',
                            help='Skip role metadata type validation when'
                                 'generating the roles_data.yaml')
        parser.add_argument('roles', nargs="+", metavar='<role>',
                            help='List of roles to use to generate the '
                                 'roles_data.yaml file for the deployment. '
                                 'NOTE: Ordering is important if no role has '
                                 'the "primary" and "controller" tags. If no '
                                 'role is tagged then the first role listed '
                                 'will be considered the primary role. This '
                                 'usually is the controller role.')
        return parser

    def _capture_output(self, filename=None):
        """Capture stdout to a file if provided"""
        if filename is not None:
            sys.stdout = open(filename, 'w')

    def _stop_capture_output(self, filename=None):
        """Stop capturing stdout to a file if provided"""
        if filename is not None:
            sys.stdout.close()

    def take_action(self, parsed_args):
        """Generate roles_data.yaml from imputed roles

        From the provided roles, validate that we have yaml files for the each
        role in our roles path and print them out concatenated together in the
        order they were provided.
        """
        self.log.debug('take_action({})'.format(parsed_args))
        roles_path = os.path.realpath(parsed_args.roles_path)
        # eliminate any dupes from the command line with an OrderedDict
        requested_roles = collections.OrderedDict.fromkeys(parsed_args.roles)
        available_roles = rolesutils.get_roles_list_from_directory(roles_path)
        rolesutils.check_role_exists(available_roles,
                                     list(requested_roles.keys()))
        self._capture_output(parsed_args.output_file)
        roles_data = rolesutils.generate_roles_data_from_directory(
            roles_path, list(requested_roles.keys()),
            parsed_args.skip_validate)
        sys.stdout.write(roles_data)
        self._stop_capture_output(parsed_args.output_file)


class RoleList(RolesBaseCommand):
    """List availables roles (DEPRECATED).

    Please use "openstack overcloud roles list" instead.
    """
    def get_parser(self, prog_name):
        parser = super(RoleList, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        self.log.debug('take_action({})'.format(parsed_args))
        self.log.warning('This command is deprecated. Please use "openstack '
                         'overcloud roles list" instead.')
        roles_path = os.path.realpath(parsed_args.roles_path)
        roles = rolesutils.get_roles_list_from_directory(roles_path)
        print('\n'.join(roles))


class RoleShow(RolesBaseCommand):
    """Show information about a given role (DEPRECATED).


    Please use "openstack overcloud roles show" intead.
    """
    def get_parser(self, prog_name):
        parser = super(RoleShow, self).get_parser(prog_name)
        parser.add_argument('role', metavar='<role>',
                            help='Role to display more information about.')
        return parser

    def take_action(self, parsed_args):
        self.log.debug('take_action({})'.format(parsed_args))
        self.log.warning('This command is deprecated. Please use "openstack '
                         'overcloud roles show" instead.')
        roles_path = os.path.realpath(parsed_args.roles_path)
        role_name = parsed_args.role
        file_path = os.path.join(roles_path, '{}.yaml'.format(role_name))
        try:
            with open(file_path, 'r') as f:
                role = rolesutils.validate_role_yaml(f)
        except IOError:
            raise NotFound("Role '{}' not found. Use 'openstack overcloud "
                           "roles list' to see the available roles.".
                           format(parsed_args.role))

        if 'name' in role:
            print('#' * 79)
            print("# Role Data for '{}'".format(role['name']))
            print('#' * 79)

        for key in sorted(role.keys()):
            print("{}:".format(key), end='')
            value = role[key]

            if isinstance(value, (list, tuple)):
                print('')
                print('\n'.join([' * {0}'.format(v) for v in value]))
            else:
                print(" '{}'".format(value))



#   Copyright 2016 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import mock
from osc_lib import exceptions

from tripleoclient import constants
from tripleoclient.tests import fakes
from tripleoclient.tests.v1.overcloud_deploy import fakes as deploy_fakes
from tripleoclient.v2 import overcloud_delete


class TestDeleteOvercloud(deploy_fakes.TestDeployOvercloud):

    def setUp(self):
        super(TestDeleteOvercloud, self).setUp()
        self.app = fakes.FakeApp()
        self.cmd = overcloud_delete.DeleteOvercloud(self.app, None)

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_plan_undeploy(self, mock_mkdir, mock_cd, mock_run_playbook):
        arglist = ["overcast", "-y"]
        verifylist = [
            ("stack", "overcast"),
            ("yes", True)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        mock_run_playbook.assert_called_once_with(
            ['cli-cleanup-ipa.yml', 'cli-overcloud-delete.yaml'],
            constants.ANSIBLE_INVENTORY,
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "stack_name": "overcast",
            },
            verbosity=3,
        )

    def test_no_confirmation(self):
        arglist = ["overcast", ]
        verifylist = [
            ("stack", "overcast"),
            ("yes", False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action, parsed_args)

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    def test_skip_ipa_cleanup(self, mock_run_playbook):
        arglist = ["overcast", "-y", "--skip-ipa-cleanup"]
        verifylist = [
            ("stack", "overcast"),
            ("yes", True),
            ("skip_ipa_cleanup", True)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        mock_run_playbook.assert_called_once_with(
            ['cli-overcloud-delete.yaml'],
            constants.ANSIBLE_INVENTORY,
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "stack_name": "overcast",
            },
            verbosity=3,
        )



#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import fixtures
import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from tripleoclient.tests.v1.test_plugin import TestPluginV1

# Load the plugin init module for the plugin list and show commands
from tripleoclient.v1 import undercloud_minion


class FakePluginV1Client(object):
    def __init__(self, **kwargs):
        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']


class TestMinionInstall(TestPluginV1):

    def setUp(self):
        super(TestMinionInstall, self).setUp()

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        # don't actually load config from ~/minion.conf
        self.mock_config_load = self.useFixture(
            fixtures.MockPatch('tripleoclient.utils.load_config'))
        # Get the command object to test
        app_args = mock.Mock()
        app_args.verbose_level = 1
        self.cmd = undercloud_minion.InstallUndercloudMinion(self.app,
                                                             app_args)

    @mock.patch('tripleoclient.v1.minion_config.prepare_minion_deploy')
    @mock.patch('tripleoclient.utils.ensure_run_as_normal_user')
    @mock.patch('tripleoclient.utils.configure_logging')
    @mock.patch('subprocess.check_call', autospec=True)
    def test_take_action(self, mock_subprocess, mock_logging, mock_usercheck,
                         mock_prepare_deploy):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        mock_prepare_deploy.return_value = ['foo']
        self.cmd.take_action(parsed_args)
        mock_prepare_deploy.assert_called_once_with(
                dry_run=False, force_stack_update=False, no_validations=False,
                verbose_level=1)
        mock_usercheck.assert_called_once()
        mock_subprocess.assert_called_with(['foo'])

    @mock.patch('tripleoclient.v1.minion_config.prepare_minion_deploy')
    @mock.patch('tripleoclient.utils.ensure_run_as_normal_user')
    @mock.patch('tripleoclient.utils.configure_logging')
    @mock.patch('subprocess.check_call', autospec=True)
    def test_take_action_dry_run(self, mock_subprocess, mock_logging,
                                 mock_usercheck, mock_prepare_deploy):
        arglist = ['--dry-run']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        mock_prepare_deploy.assert_called_once_with(
                dry_run=True, force_stack_update=False, no_validations=True,
                verbose_level=1)
        mock_usercheck.assert_called_once()
        self.assertCountEqual(mock_subprocess.call_args_list, [])


class TestMinionUpgrade(TestPluginV1):

    def setUp(self):
        super(TestMinionUpgrade, self).setUp()

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        # don't actually load config from ~/minion.conf
        self.mock_config_load = self.useFixture(
            fixtures.MockPatch('tripleoclient.utils.load_config'))
        # Get the command object to test
        app_args = mock.Mock()
        app_args.verbose_level = 1
        self.cmd = undercloud_minion.UpgradeUndercloudMinion(self.app,
                                                             app_args)

    @mock.patch('tripleoclient.v1.minion_config.prepare_minion_deploy')
    @mock.patch('tripleoclient.utils.ensure_run_as_normal_user')
    @mock.patch('tripleoclient.utils.configure_logging')
    @mock.patch('subprocess.check_call', autospec=True)
    def test_take_action(self, mock_subprocess, mock_logging, mock_usercheck,
                         mock_prepare_deploy):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        mock_prepare_deploy.return_value = ['foo']
        self.cmd.take_action(parsed_args)
        mock_prepare_deploy.assert_called_once_with(
                force_stack_update=False, no_validations=False, upgrade=True,
                verbose_level=1, yes=False)
        mock_usercheck.assert_called_once()
        mock_subprocess.assert_called_with(['foo'])

    @mock.patch('tripleoclient.v1.minion_config.prepare_minion_deploy')
    @mock.patch('tripleoclient.utils.ensure_run_as_normal_user')
    @mock.patch('tripleoclient.utils.configure_logging')
    @mock.patch('subprocess.check_call', autospec=True)
    def test_take_action_yes(self, mock_subprocess, mock_logging,
                             mock_usercheck, mock_prepare_deploy):
        arglist = ['--yes']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        mock_prepare_deploy.return_value = ['foo']
        self.cmd.take_action(parsed_args)
        mock_prepare_deploy.assert_called_once_with(
                force_stack_update=False, no_validations=False, upgrade=True,
                verbose_level=1, yes=True)
        mock_usercheck.assert_called_once()
        mock_subprocess.assert_called_with(['foo'])



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from heatclient.common import event_utils
from tripleo_common.actions import scale

from tripleoclient import utils
from tripleoclient.workflows import deployment


def scale_down(log, clients, stack, nodes, timeout=None, verbosity=0,
               connection_timeout=None):
    """Unprovision and deletes overcloud nodes from a heat stack.

    :param log: Logging object
    :type log: Object

    :param clients: Application client object.
    :type clients: Object

    :param stack: Heat Stack object
    :type stack: Object

    :param nodes: List of nodes to delete. If the node UUID is used the
                  UUID will be used to lookup the node name before being
                  passed through to the cleanup playbook.
    :type nodes: List

    :param timeout: Timeout to use when deleting nodes. If timeout is None
                    it will be set to 240 minutes.
    :type timeout: Integer

    :param verbosity: Verbosity level
    :type verbosity: Integer

    :param connection_timeout: Ansible connection timeout in seconds.
    :type connection_timeout: Integer
    """

    if not timeout:
        timeout = 240

    limit_list = list()
    for node in nodes:
        try:
            _node = clients.compute.servers.get(node)
            limit_list.append(_node.name)
        except Exception:
            limit_list.append(node)

    if limit_list:
        limit_list = ':'.join(limit_list)
    else:
        limit_list = None

    deployment.config_download(
        log=log,
        clients=clients,
        stack=stack,
        timeout=connection_timeout,
        ansible_playbook_name='scale_playbook.yaml',
        limit_hosts=limit_list,
        verbosity=verbosity,
        deployment_timeout=timeout
    )
    events = event_utils.get_events(clients.orchestration,
                                    stack_id=stack.stack_name,
                                    event_args={'sort_dir': 'desc',
                                                'limit': 1})
    marker = events[0].id if events else None

    print('Running scale down')
    context = clients.tripleoclient.create_mistral_context()
    scale_down_action = scale.ScaleDownAction(nodes=nodes, timeout=timeout,
                                              container=stack.stack_name)
    scale_down_action.run(context=context)
    utils.wait_for_stack_ready(
        orchestration_client=clients.orchestration,
        stack_name=stack.stack_name,
        action='UPDATE',
        marker=marker
    )



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from tripleoclient.tests import fakes
from tripleoclient.workflows import baremetal


class TestBaremetalWorkflows(fakes.FakePlaybookExecution):

    def setUp(self):
        super(TestBaremetalWorkflows, self).setUp()

        self.app.client_manager.workflow_engine = self.workflow = mock.Mock()
        self.tripleoclient = mock.Mock()
        self.app.client_manager.tripleoclient = self.tripleoclient

        self.mock_playbook = mock.patch(
            'tripleoclient.utils.run_ansible_playbook',
            autospec=True
        )
        self.mock_playbook.start()
        self.addCleanup(self.mock_playbook.stop)

    def test_register_or_update_success(self):
        self.assertEqual(baremetal.register_or_update(
            self.app.client_manager,
            nodes_json=[],
            instance_boot_option='local'
        ), [mock.ANY])

    def test_provide_success(self):
        baremetal.provide(self.app.client_manager, node_uuids=[])

    def test_format_errors(self):
        payload = {'message': [{'result': 'Error1a\nError1b'},
                               {'result': 'Error2a\nError2b\n'}]}

        error_string = baremetal._format_errors(payload)
        self.assertEqual(error_string, "Error1b\nError2b")

    def test_introspect_success(self):
        baremetal.introspect(self.app.client_manager, node_uuids=[],
                             run_validations=True, concurrency=20,
                             node_timeout=1200, max_retries=1,
                             retry_timeout=120)

    def test_introspect_manageable_nodes_success(self):
        baremetal.introspect_manageable_nodes(
            self.app.client_manager, run_validations=False, concurrency=20,
            node_timeout=1200, max_retries=1, retry_timeout=120,
        )

    def test_provide_manageable_nodes_success(self):
        baremetal.provide_manageable_nodes(
            self.app.client_manager
        )

    def test_configure_success(self):
        baremetal.configure(self.app.client_manager, node_uuids=[])

    def test_configure_manageable_nodes_success(self):
        baremetal.configure_manageable_nodes(self.app.client_manager)

    def test_clean_nodes_success(self):
        baremetal.clean_nodes(node_uuids=[])

    def test_clean_manageable_nodes_success(self):
        baremetal.clean_manageable_nodes(
            self.app.client_manager
        )



#   Copyright 2016 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import json
import logging
import netaddr
import os
import pwd
import re
import shutil
import six
import subprocess
import sys
import tarfile
import tempfile
import time
import traceback
import yaml

from cliff import command
from datetime import datetime
from heatclient.common import template_utils
from osc_lib.i18n import _
from six.moves import configparser

from tripleoclient import constants
from tripleoclient import exceptions
from tripleoclient import heat_launcher
from tripleoclient import utils

from tripleo_common import constants as tc_constants
from tripleo_common.image import kolla_builder
from tripleo_common.utils import parameters
from tripleo_common.utils import passwords as password_utils

# For ansible download and config generation
from tripleo_common.actions import ansible
from tripleo_common.inventory import TripleoInventory
from tripleo_common.utils import config

DEPLOY_FAILURE_MESSAGE = """
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

Deployment Failed!

ERROR: Heat log files: {0}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
"""
DEPLOY_COMPLETION_MESSAGE = """
########################################################

Deployment successful!

########################################################
"""
OUTPUT_ONLY_COMPLETION_MESSAGE = """
########################################################

Deployment information successfully generated!

########################################################
"""
STANDALONE_COMPLETION_MESSAGE = """
##########################################################

Useful files:

The clouds.yaml file is at {0}

Use "export OS_CLOUD=standalone" before running the
openstack command.

##########################################################
"""


class Deploy(command.Command):
    """Deploy containerized Undercloud"""

    log = logging.getLogger(__name__ + ".Deploy")
    auth_required = False
    heat_pid = None
    tht_render = None
    output_dir = None
    tmp_ansible_dir = None
    stack_update_mark = None
    stack_action = 'CREATE'
    deployment_user = None
    ansible_dir = None
    python_version = sys.version_info[0]
    ansible_playbook_cmd = "ansible-playbook"
    python_cmd = "python{}".format(python_version)

    def _is_undercloud_deploy(self, parsed_args):
        role = parsed_args.standalone_role
        stack = parsed_args.stack
        return (role in ['Undercloud', 'UndercloudMinion'] and
                stack in ['undercloud', 'minion'])

    def _run_preflight_checks(self, parsed_args):
        """Run preflight deployment checks

        Perform any pre-deployment checks that we want to run when deploying
        standalone deployments. This is skipped when in output only mode or
        when used with an undercloud. The undercloud has it's own set of
        deployment preflight requirements.

        :param parsed_args: parsed arguments from the cli
        """
        # we skip preflight checks for output only and undercloud
        if parsed_args.output_only or self._is_undercloud_deploy(parsed_args):
            return

        # in standalone we don't want to fixup the /etc/hosts as we'll be
        # managing that elsewhere during the deployment
        utils.check_hostname(fix_etc_hosts=False, logger=self.log)

        # Users can use http_proxy and https_proxy as part of the deployment,
        # however we need localhost to not be proxied because we use it to talk
        # to our heat api.
        utils.check_env_for_proxy(no_proxy_hosts=['127.0.0.1'])

    # NOTE(cjeanner) Quick'n'dirty way before we have proper
    # escalation support through oslo.privsep
    def _set_data_rights(self, file_name, user=None,
                         mode=0o600):

        u = user or self.deployment_user
        u_flag = None
        f_flag = None
        if u:
            if os.path.exists(file_name):
                try:
                    pwd.getpwnam(u)
                    cmd = 'sudo chown -R %s %s' % (u, file_name)
                    subprocess.check_call(cmd.split())
                except KeyError:
                    u_flag = 'Unknown'
            else:
                f_flag = "Absent"
        else:
            u_flag = 'Undefined'

        if u_flag:
            self.log.warning(_('%(u_f)s user "%(u)s". You might need to '
                             'manually set ownership after the deploy')
                             % {'u_f': u_flag, 'u': user})
        if f_flag:
            self.log.warning(_('%(f)s file is %(f_f)s.')
                             % {'f': file_name, 'f_f': f_flag})
        else:
            os.chmod(file_name, mode)

    def _set_stack_action(self, parsed_args):
        """Set the stack action for deployment"""
        # Prepare the heat stack action we want to start deployment with
        if ((os.path.isfile(self.stack_update_mark) or
           parsed_args.force_stack_update) and
           not parsed_args.force_stack_create):
            self.stack_action = 'UPDATE'

        self.log.warning(
            _('The heat stack {0} action is {1}').format(
                parsed_args.stack, self.stack_action))

    def _get_roles_file_path(self, parsed_args):
        """Return roles_file for the deployment"""
        if not parsed_args.roles_file:
            roles_file = os.path.join(parsed_args.templates,
                                      constants.UNDERCLOUD_ROLES_FILE)
        else:
            roles_file = parsed_args.roles_file
        return roles_file

    def _get_networks_file_path(self, parsed_args):
        """Return networks_file for the deployment"""
        if not parsed_args.networks_file:
            return os.path.join(parsed_args.templates,
                                constants.STANDALONE_NETWORKS_FILE)
        else:
            return parsed_args.networks_file

    def _get_plan_env_file_path(self, parsed_args):
        """Return plan_environment_file for the deployment"""
        if not parsed_args.plan_environment_file:
            plan_env = os.path.join(parsed_args.templates,
                                    constants.PLAN_ENVIRONMENT)
        else:
            plan_env = parsed_args.plan_environment_file
        return plan_env

    def _get_primary_role_name(self, roles_file_path, templates):
        """Return the primary role name"""
        roles_data = utils.fetch_roles_file(
            roles_file_path, templates)
        if not roles_data:
            # TODO(aschultz): should this be Undercloud instead?
            return 'Controller'

        for r in roles_data:
            if 'tags' in r and 'primary' in r['tags']:
                return r['name']
        self.log.warning('No primary role found in roles_data, using '
                         'first defined role')
        return roles_data[0]['name']

    def _get_tar_filename(self, stack_name='undercloud'):
        """Return tarball name for the install artifacts"""
        return '%s/%s-install-%s.tar.bzip2' % \
               (self.output_dir, stack_name,
                datetime.utcnow().strftime('%Y%m%d%H%M%S'))

    def _create_install_artifact(self, user, stack_name='undercloud'):
        """Create a tarball of the temporary folders used"""
        self.log.debug(_("Preserving deployment artifacts"))

        def remove_output_dir(info):
            """Tar filter to remove output dir from path"""
            # leading path to tar is home/stack/ rather than /home/stack
            leading_path = self.output_dir[1:] + '/'
            info.name = info.name.replace(leading_path, '')
            return info

        # tar up working data and put in
        # output_dir/undercloud-install-TS.tar.bzip2
        # if the stack name is "undercloud".
        tar_filename = self._get_tar_filename(stack_name)
        try:
            tf = tarfile.open(tar_filename, 'w:bz2')
            tf.add(self.tht_render, recursive=True, filter=remove_output_dir)
            tf.add(self.tmp_ansible_dir, recursive=True,
                   filter=remove_output_dir)
            tf.close()
        except Exception as ex:
            msg = _("Unable to create artifact tarball, %s") % ex.message
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)
        # TODO(cjeanner) drop that once using oslo.privsep
        self._set_data_rights(tar_filename, user=user)
        return tar_filename

    def _create_persistent_dirs(self):
        """Creates temporary working directories"""
        utils.makedirs(constants.STANDALONE_EPHEMERAL_STACK_VSTATE)

    def _create_working_dirs(self, stack_name='undercloud'):
        """Creates temporary working directories"""
        if self.output_dir:
            utils.makedirs(self.output_dir)
        if not self.tht_render:
            self.tht_render = os.path.join(self.output_dir,
                                           'tripleo-heat-installer-templates')
            # Clear dir since we're using a static name and shutils.copytree
            # needs the folder to not exist. We'll generate the
            # contents each time. This should clear the folder on the first
            # run of this function.
            shutil.rmtree(self.tht_render, ignore_errors=True)
        if not self.tmp_ansible_dir:
            self.tmp_ansible_dir = tempfile.mkdtemp(
                prefix=stack_name + '-ansible-', dir=self.output_dir)

    def _populate_templates_dir(self, source_templates_dir,
                                stack_name='undercloud'):
        """Creates template dir with templates

        * Copy --templates content into a working dir
          created as 'output_dir/tripleo-heat-installer-templates'.

        :param source_templates_dir: string to a directory containing our
                                     source templates
        """
        self._create_working_dirs(stack_name)
        if not os.path.exists(source_templates_dir):
            raise exceptions.NotFound("%s templates directory does not exist "
                                      "or permission denied" %
                                      source_templates_dir)
        if not os.path.exists(self.tht_render):
            shutil.copytree(source_templates_dir, self.tht_render,
                            symlinks=True)

    def _set_default_plan(self):
        """Populate default plan-environment.yaml."""
        if not os.path.isfile(os.path.join(self.tht_render,
                              'plan-environment.yaml')):
            shutil.copy(os.path.join(self.tht_render, 'plan-samples',
                        'openstack', 'plan-environment.yaml'), self.tht_render)

    def _cleanup_working_dirs(self, cleanup=False, user=None):
        """Cleanup temporary working directories

        :param cleanup: Set to true if you DO want to cleanup the dirs
        """
        if cleanup:
            if self.tht_render and os.path.exists(self.tht_render):
                shutil.rmtree(self.tht_render, ignore_errors=True)

            self.tht_render = None
            if self.tmp_ansible_dir and os.path.exists(self.tmp_ansible_dir):
                shutil.rmtree(self.tmp_ansible_dir)
                self.tmp_ansible_dir = None
        else:
            self.log.warning(_("Not cleaning working directory %s")
                             % self.tht_render)
            # TODO(cjeanner) drop that once using oslo.privsep
            self._set_data_rights(self.tht_render, user=user, mode=0o700)
            self.log.warning(_("Not cleaning ansible directory %s")
                             % self.tmp_ansible_dir)
            # TODO(cjeanner) drop that once using oslo.privsep
            self._set_data_rights(self.tmp_ansible_dir, user=user, mode=0o700)

    def _configure_puppet(self):
        self.log.info(_('Configuring puppet modules symlinks ...'))
        utils.bulk_symlink(self.log, constants.TRIPLEO_PUPPET_MODULES,
                           constants.PUPPET_MODULES,
                           constants.PUPPET_BASE)

    def _update_passwords_env(self, output_dir, user, upgrade=None,
                              passwords=None, stack_name='undercloud'):
        pw_file = os.path.join(output_dir,
                               'tripleo-' + stack_name + '-passwords.yaml')
        undercloud_pw_file = os.path.join(output_dir,
                                          stack_name + '-passwords.conf')

        # Generated passwords take the lowest precedence, allowing
        # custom overrides
        stack_env = {'parameter_defaults': {}}
        stack_env['parameter_defaults'] = password_utils.generate_passwords(
            stack_env=stack_env)
        if os.path.exists(pw_file):
            with open(pw_file) as pf:
                stack_env['parameter_defaults'].update(
                    yaml.safe_load(pf.read())['parameter_defaults'])

        if upgrade:
            # Getting passwords that were managed by instack-undercloud so
            # we can upgrade to a containerized undercloud and keep old
            # passwords.
            legacy_env = {}
            if os.path.exists(undercloud_pw_file):
                config = configparser.ConfigParser()
                config.read(undercloud_pw_file)
                for k, v in config.items('auth'):
                    # Manage exceptions
                    if k == 'undercloud_db_password':
                        k = 'MysqlRootPassword'
                    elif k == 'undercloud_rabbit_username':
                        k = 'RpcUserName'
                    elif k == 'undercloud_rabbit_password':
                        try:
                            # NOTE(aschultz): Only save rabbit password to rpc
                            # if it's not already defined for the upgrade case.
                            # The passwords are usually different so we don't
                            # want to overwrite it if it already exists because
                            # we'll end up rewriting the passwords later and
                            # causing problems.
                            config.get('auth', 'undercloud_rpc_password')
                        except Exception:
                            legacy_env['RpcPassword'] = v
                        k = 'RabbitPassword'
                    elif k == 'undercloud_rabbit_cookie':
                        k = 'RabbitCookie'
                    elif k == 'undercloud_heat_encryption_key':
                        k = 'HeatAuthEncryptionKey'
                    elif k == 'undercloud_libvirt_tls_password':
                        k = 'LibvirtTLSPassword'
                    elif k == 'undercloud_ha_proxy_stats_password':
                        k = 'HAProxyStatsPassword'
                    else:
                        k = ''.join(i.capitalize() for i in k.split('_')[1:])
                    legacy_env[k] = v

            # Get the keystone keys before upgrade
            keystone_fernet_repo = '/etc/keystone/fernet-keys/'
            keystone_credential_repo = '/etc/keystone/credential-keys/'
            self._set_data_rights('/etc/keystone', user=user)

            for key_index in range(0, 2):
                file_name = keystone_credential_repo + str(key_index)
                key = 'KeystoneCredential' + str(key_index)
                if os.path.exists(file_name):
                    with open(file_name, 'r') as file_content:
                        content = file_content.read()
                        legacy_env[key] = content

            fernet_keys = {}
            file_count = 0
            if os.path.exists(keystone_fernet_repo):
                file_count = len(os.listdir(keystone_fernet_repo))

            for key_index in range(0, file_count):
                file_name = keystone_fernet_repo + str(key_index)
                if os.path.exists(file_name):
                    with open(file_name, 'r') as file_content:
                        content = file_content.read()
                        fernet_keys[file_name] = {'content': content}
            if fernet_keys:
                legacy_env['KeystoneFernetKeys'] = fernet_keys

            # Override with picked legacy instack-undercloud values
            stack_env['parameter_defaults'].update(legacy_env)

        if passwords:
            # These passwords are the DefaultPasswords so we only
            # update if they don't already exist in stack_env
            for p, v in passwords.items():
                if p not in stack_env['parameter_defaults']:
                    stack_env['parameter_defaults'][p] = v

        # Write out the password file in yaml for heat.
        # This contains sensitive data so ensure it's not world-readable
        with open(pw_file, 'w') as pf:
            yaml.safe_dump(stack_env, pf, default_flow_style=False)
        # TODO(cjeanner) drop that once using oslo.privsep
        # Do not forget to re-add os.chmod 0o600 on that one!
        self._set_data_rights(pw_file, user=user)
        # Write out an instack undercloud compatible version.
        # This contains sensitive data so ensure it's not world-readable
        with open(undercloud_pw_file, 'w') as pf:
            pf.write('[auth]\n')
            for p, v in stack_env['parameter_defaults'].items():
                if 'Password' in p or 'Token' in p or p.endswith('Kek'):
                    # Convert camelcase from heat templates into the underscore
                    # format used by instack undercloud.
                    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', p)
                    pw_key = re.sub('([a-z0-9])([A-Z])',
                                    r'\1_\2', s1).lower()
                    pf.write('undercloud_%s: %s\n' % (pw_key, v))

        # TODO(cjeanner) drop that once using oslo.privsep
        # Do not forget to re-add os.chmod 0o600 on that one!
        self._set_data_rights(undercloud_pw_file, user=user)

        return pw_file

    def _generate_hosts_parameters(self, parsed_args, p_ip):
        hostname = utils.get_short_hostname()
        domain = parsed_args.local_domain

        data = {
            'CloudName': p_ip,
            'CloudDomain': domain,
            'CloudNameInternal': '%s.internalapi.%s' % (hostname, domain),
            'CloudNameStorage': '%s.storage.%s' % (hostname, domain),
            'CloudNameStorageManagement': ('%s.storagemgmt.%s'
                                           % (hostname, domain)),
            'CloudNameCtlplane': '%s.ctlplane.%s' % (hostname, domain),
        }
        return data

    def _generate_portmap_parameters(self, ip_addr, ip_nw, ctlplane_vip_addr,
                                     public_vip_addr, stack_name='Undercloud',
                                     role_name='Undercloud'):
        hostname = utils.get_short_hostname()

        # in order for deployed server network information to match correctly,
        # we need to ensure the HostnameMap matches our hostname
        hostname_map_name = "%s-%s-0" % (stack_name.lower(), role_name.lower())
        data = {
            'HostnameMap': {
                hostname_map_name: '%s' % hostname
            },
            # The settings below allow us to inject a custom public
            # VIP. This requires use of the generated
            # ../network/ports/external_from_pool.yaml resource in t-h-t.
            'IPPool': {
                'external': [public_vip_addr]
            },
            'ExternalNetCidr': '%s/%s' % (public_vip_addr, ip_nw.prefixlen),
            # This requires use of the
            # ../deployed-server/deployed-neutron-port.yaml resource in t-h-t
            # We use this for the control plane VIP and also via
            # the environments/deployed-server-noop-ctlplane.yaml
            # for the server IP itself
            'DeployedServerPortMap': {
                ('%s-ctlplane' % hostname): {
                    'fixed_ips': [{'ip_address': ip_addr}],
                    'subnets': [{'cidr': str(ip_nw.cidr)}],
                    'network': {'tags': [str(ip_nw.cidr)]}
                },
                'control_virtual_ip': {
                    'fixed_ips': [{'ip_address': ctlplane_vip_addr}],
                    'subnets': [{'cidr': str(ip_nw.cidr)}],
                    'network': {'tags': [str(ip_nw.cidr)]}
                },
                'public_virtual_ip': {
                    'fixed_ips': [{'ip_address': public_vip_addr}],
                    'subnets': [{'cidr': str(ip_nw.cidr)}],
                    'network': {'tags': [str(ip_nw.cidr)]}
                }
            }
        }
        return data

    def _kill_heat(self, parsed_args):
        """Tear down heat installer and temp files

        Kill the heat launcher/installer process.
        Teardown temp files created in the deployment process,
        when cleanup is requested.

        """
        if self.heat_pid:
            self.heat_launch.kill_heat(self.heat_pid)
            pid, ret = os.waitpid(self.heat_pid, 0)
            self.heat_pid = None

    def _launch_heat(self, parsed_args):
        # we do this as root to chown config files properly for docker, etc.
        if parsed_args.heat_native is not None and \
                parsed_args.heat_native.lower() == "false":
            self.heat_launch = heat_launcher.HeatContainerLauncher(
                parsed_args.heat_api_port,
                parsed_args.heat_container_image,
                parsed_args.heat_user)
        else:
            self.heat_launch = heat_launcher.HeatNativeLauncher(
                parsed_args.heat_api_port,
                parsed_args.heat_container_image,
                parsed_args.heat_user)

        # NOTE(dprince): we launch heat with fork exec because
        # we don't want it to inherit our args. Launching heat
        # as a "library" would be cool... but that would require
        # more refactoring. It runs a single process and we kill
        # it always below.
        self.heat_pid = os.fork()
        if self.heat_pid == 0:
            if parsed_args.heat_native is not None and \
                    parsed_args.heat_native.lower() == "true":
                try:
                    uid = pwd.getpwnam(parsed_args.heat_user).pw_uid
                    gid = pwd.getpwnam(parsed_args.heat_user).pw_gid
                except KeyError:
                    msg = _(
                        "Please create a %s user account before "
                        "proceeding.") % parsed_args.heat_user
                    self.log.error(msg)
                    raise exceptions.DeploymentError(msg)
                os.setgid(gid)
                os.setuid(uid)
            self.heat_launch.heat_db_sync()
            # Exec() never returns.
            self.heat_launch.launch_heat()

        # NOTE(dprince): we use our own client here because we set
        # auth_required=False above because keystone isn't running when this
        # command starts
        tripleoclients = self.app.client_manager.tripleoclient
        orchestration_client = \
            tripleoclients.local_orchestration(parsed_args.heat_api_port)

        return orchestration_client

    def _normalize_user_templates(self, user_tht_root, tht_root, env_files=[]):
        """copy environment files into tht render path

        This assumes any env file that includes user_tht_root has already
        been copied into tht_root.

        :param user_tht_root: string path to the user's template dir
        :param tht_root: string path to our deployed tht_root
        :param env_files: list of paths to environment files
        :return list of absolute pathed environment files that exist in
                tht_root
        """
        environments = []
        # normalize the user template path to ensure it doesn't have a trailing
        # slash
        user_tht = os.path.abspath(user_tht_root)
        for env_path in env_files:
            self.log.debug("Processing file %s" % env_path)
            abs_env_path = os.path.abspath(env_path)
            if (abs_env_path.startswith(user_tht_root) and
                    ((user_tht + '/') in env_path or
                     (user_tht + '/') in abs_env_path or
                     user_tht == abs_env_path or
                     user_tht == env_path)):
                # file is in tht and will be copied, so just update path
                new_env_path = env_path.replace(user_tht + '/',
                                                tht_root + '/')
                self.log.debug("Redirecting %s to %s"
                               % (abs_env_path, new_env_path))
                environments.append(new_env_path)
            elif abs_env_path.startswith(tht_root):
                self.log.debug("File already in tht_root %s")
                environments.append(abs_env_path)
            else:
                self.log.debug("File outside of tht_root %s, copying in")
                # file is outside of THT, just copy it in
                # TODO(aschultz): probably shouldn't be flattened?
                target_dest = os.path.join(tht_root,
                                           os.path.basename(abs_env_path))
                if os.path.exists(target_dest):
                    raise exceptions.DeploymentError("%s already exists, "
                                                     "please rename the "
                                                     "file to something else"
                                                     % target_dest)
                shutil.copy(abs_env_path, tht_root)
                environments.append(target_dest)
        return environments

    def _load_user_params(self, user_environments):
        user_params = {}
        for env_file in user_environments:
            # undercloud heat stack virtual state tracking is'nt available yet
            if env_file.endswith('undercloud-stack-vstate-dropin.yaml'):
                continue

            with open(env_file, 'r') as f:
                data = yaml.safe_load(f.read())

            if data is None or data.get('parameter_defaults') is None:
                continue

            for k, v in data.get('parameter_defaults', {}).items():
                user_params[k] = v

        return user_params

    def _setup_heat_environments(self, roles_file_path, networks_file_path,
                                 parsed_args):
        """Process tripleo heat templates with jinja and deploy into work dir

        * Process j2/install additional templates there
        * Return the environments list for futher processing as a new base.

        The first two items are reserved for the
        overcloud-resource-registry-puppet.yaml and passwords files.
        """

        self.log.warning(_("** Handling template files **"))
        env_files = []

        # TODO(aschultz): in overcloud deploy we have a --environments-dir
        # we might want to handle something similar for this
        # (shardy) alternatively perhaps we should rely on the plan-environment
        # environments list instead?
        if parsed_args.environment_files:
            env_files.extend(parsed_args.environment_files)

        # ensure any user provided templates get copied into tht_render
        user_environments = self._normalize_user_templates(
            parsed_args.templates, self.tht_render, env_files)

        # generate jinja templates by its work dir location
        self.log.debug(_("Using roles file %s") % roles_file_path)
        process_templates = os.path.join(parsed_args.templates,
                                         'tools/process-templates.py')
        args = [self.python_cmd, process_templates, '--roles-data',
                roles_file_path, '--network-data', networks_file_path,
                '--output-dir', self.tht_render]
        if utils.run_command_and_log(self.log, args, cwd=self.tht_render) != 0:
            # TODO(aschultz): improve error messaging
            msg = _("Problems generating templates.")
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)

        # NOTE(aschultz): the next set of environment files are system included
        # so we have to include them at the front of our environment list so a
        # user can override anything in them.

        # Include any environments from the plan-environment.yaml
        plan_env_path = utils.rel_or_abs_path(
            self._get_plan_env_file_path(parsed_args), self.tht_render)
        with open(plan_env_path, 'r') as f:
            plan_env_data = yaml.safe_load(f)
        environments = [utils.rel_or_abs_path(e.get('path'), self.tht_render)
                        for e in plan_env_data.get('environments', {})]

        # this will allow the user to overwrite passwords with custom envs
        # or pick instack legacy passwords as is, if upgrading from instack
        pw_file = self._update_passwords_env(
            output_dir=self.output_dir,
            user=parsed_args.deployment_user,
            upgrade=parsed_args.upgrade,
            stack_name=parsed_args.stack.lower(),
        )
        environments.append(pw_file)

        # use deployed-server because we run os-collect-config locally
        deployed_server_env = os.path.join(
            self.tht_render, 'environments',
            'deployed-server-noop-ctlplane.yaml')
        environments.append(deployed_server_env)

        self.log.info(_("Deploying templates in the directory {0}").format(
            os.path.abspath(self.tht_render)))

        maps_file = os.path.join(self.tht_render,
                                 'tripleoclient-hosts-portmaps.yaml')
        ip_nw = netaddr.IPNetwork(parsed_args.local_ip)
        ip = str(ip_nw.ip)

        if parsed_args.control_virtual_ip:
            c_ip = parsed_args.control_virtual_ip
        else:
            c_ip = ip

        if parsed_args.public_virtual_ip:
            p_ip = parsed_args.public_virtual_ip
        else:
            p_ip = ip

        tmp_env = self._generate_hosts_parameters(parsed_args, p_ip)
        tmp_env.update(self._generate_portmap_parameters(
            ip, ip_nw, c_ip, p_ip,
            stack_name=parsed_args.stack,
            role_name=self._get_primary_role_name(
                roles_file_path, parsed_args.templates)))

        user_params = self._load_user_params(user_environments)
        host_routes = user_params.get('ControlPlaneStaticRoutes', [])
        mtu = user_params.get('InterfaceLocalMtu', 1500)
        tmp_env.update(
            {
                'CtlplaneNetworkAttributes': {
                    'network': {
                        'mtu': mtu,
                    },
                    'subnets': {
                        'ctlplane-subnet': {
                            'cidr': str(ip_nw.cidr),
                            'host_routes': host_routes,
                        }
                    }
                }
            }
        )

        with open(maps_file, 'w') as env_file:
            yaml.safe_dump({'parameter_defaults': tmp_env}, env_file,
                           default_flow_style=False)
        environments.append(maps_file)

        # NOTE(aschultz): this doesn't get copied into tht_root but
        # we always include the hieradata override stuff last.
        if parsed_args.hieradata_override:
            environments.append(self._process_hieradata_overrides(
                parsed_args.hieradata_override,
                parsed_args.standalone_role,
                parsed_args.stack.lower()))

        # Create a persistent drop-in file to indicate the stack
        # virtual state changes
        stack_vstate_dropin = os.path.join(self.tht_render,
                                           '%s-stack-vstate-dropin.yaml' %
                                           parsed_args.stack)
        with open(stack_vstate_dropin, 'w') as dropin_file:
            yaml.safe_dump(
                {'parameter_defaults': {
                    'RootStackName': parsed_args.stack.lower(),
                    'StackAction': self.stack_action,
                    'DeployIdentifier': int(time.time())}},
                dropin_file, default_flow_style=False)
        environments.append(stack_vstate_dropin)

        return environments + user_environments

    def _prepare_container_images(self, env, roles_data):
        image_params = kolla_builder.container_images_prepare_multi(
            env, roles_data, dry_run=True)

        # use setdefault to ensure every needed image parameter is
        # populated without replacing user-set values
        if image_params:
            pd = env.get('parameter_defaults', {})
            for k, v in image_params.items():
                pd.setdefault(k, v)

    def _deploy_tripleo_heat_templates(self, orchestration_client,
                                       parsed_args):
        """Deploy the fixed templates in TripleO Heat Templates"""
        roles_file_path = self._get_roles_file_path(parsed_args)
        networks_file_path = self._get_networks_file_path(parsed_args)

        # sets self.tht_render to the working dir with deployed templates
        environments = self._setup_heat_environments(
            roles_file_path, networks_file_path, parsed_args)

        # rewrite paths to consume t-h-t env files from the working dir
        self.log.debug(_("Processing environment files %s") % environments)
        env_files, env = utils.process_multiple_environments(
            environments, self.tht_render, parsed_args.templates,
            cleanup=parsed_args.cleanup)

        roles_data = utils.fetch_roles_file(
            roles_file_path, parsed_args.templates)
        to_remove = set()
        for key, value in env.get('resource_registry', {}).items():
            if (key.startswith('OS::TripleO::Services::') and
                    value == 'OS::Heat::None'):
                to_remove.add(key)
        if to_remove:
            for role in roles_data:
                for service in to_remove:
                    try:
                        role.get('ServicesDefault', []).remove(service)
                    except ValueError:
                        pass
            self.log.info('Removing unused services, updating roles')
            # This will clean up the directory and set it up again
            self.tht_render = None
            self._populate_templates_dir(parsed_args.templates,
                                         parsed_args.stack.lower())
            roles_file_path = os.path.join(
                self.tht_render, 'roles-data-override.yaml')
            with open(roles_file_path, "w") as f:
                f.write(yaml.safe_dump(roles_data))
            # Redo the dance
            environments = self._setup_heat_environments(
                roles_file_path, networks_file_path, parsed_args)
            env_files, env = utils.process_multiple_environments(
                environments, self.tht_render, parsed_args.templates,
                cleanup=parsed_args.cleanup)

        if not parsed_args.disable_container_prepare:
            self._prepare_container_images(env, roles_data)
        parameters.convert_docker_params(env)

        self.log.debug(_("Getting template contents"))
        template_path = os.path.join(self.tht_render, 'overcloud.yaml')
        template_files, template = \
            template_utils.get_template_contents(template_path)

        files = dict(list(template_files.items()) + list(env_files.items()))

        stack_name = parsed_args.stack

        self.log.debug(_("Deploying stack: %s") % stack_name)
        self.log.debug(_("Deploying template: %s") % template)
        self.log.debug(_("Deploying environment: %s") % env)
        self.log.debug(_("Deploying files: %s") % files)

        stack_args = {
            'stack_name': stack_name,
            'template': template,
            'environment': env,
            'files': files,
        }

        if parsed_args.timeout:
            stack_args['timeout_mins'] = parsed_args.timeout

        self.log.warning(_("** Performing Heat stack create.. **"))
        stack = orchestration_client.stacks.create(**stack_args)
        if not stack:
            msg = _('The ephemeral Heat stack could not be created, please '
                    'check logs in /var/log/heat-launcher and/or any '
                    'possible misconfiguration.')
            raise exceptions.DeploymentError(msg)

        stack_id = stack['stack']['id']
        return "%s/%s" % (stack_name, stack_id)

    def _download_ansible_playbooks(self, client, stack_name,
                                    tripleo_role_name='Standalone',
                                    python_interpreter=sys.executable):
        stack_config = config.Config(client)
        self._create_working_dirs(stack_name.lower())

        self.log.warning(_('** Downloading {0} ansible.. **').format(
            stack_name))
        # python output buffering is making this seem to take forever..
        sys.stdout.flush()
        stack_config.write_config(stack_config.fetch_config(stack_name),
                                  stack_name,
                                  self.tmp_ansible_dir)

        inventory = TripleoInventory(
            hclient=client,
            plan_name=stack_name,
            ansible_ssh_user='root')

        inv_path = os.path.join(self.tmp_ansible_dir, 'inventory.yaml')
        extra_vars = {
            tripleo_role_name: {
                'ansible_connection': 'local',
                'ansible_python_interpreter': python_interpreter,
                }
            }

        inventory.write_static_inventory(inv_path, extra_vars)

        self.log.info(_('** Downloaded {0} ansible to {1} **').format(
                      stack_name, self.tmp_ansible_dir))
        sys.stdout.flush()
        return self.tmp_ansible_dir

    def _download_stack_outputs(self, client, stack_name):
        stack = utils.get_stack(client, stack_name)
        output_file = 'tripleo-{}-outputs.yaml'.format(stack_name)
        endpointmap_file = os.path.join(self.output_dir, output_file)

        outputs = {}
        endpointmap = utils.get_endpoint_map(stack)
        if endpointmap:
            outputs['EndpointMapOverride'] = endpointmap

        allnodescfg = utils.get_stack_output_item(stack, 'AllNodesConfig')
        if allnodescfg:
            outputs['AllNodesExtraMapData'] = allnodescfg

        hosts = utils.get_stack_output_item(stack, 'HostsEntry')
        if hosts:
            outputs['ExtraHostFileEntries'] = hosts

        globalcfg = utils.get_stack_output_item(stack, 'GlobalConfig')
        if globalcfg:
            # unfortunately oslo messaging doesn't use the standard endpoint
            # configurations so we need to build out the node name vars and
            # we want to grab it for the undercloud for use by a minion.
            # The same is true for memcached as well.
            rolenet = utils.get_stack_output_item(stack, 'RoleNetIpMap')
            if 'Undercloud' in rolenet:
                name = rolenet['Undercloud']['ctlplane']
                globalcfg['oslo_messaging_rpc_node_names'] = [name]
                globalcfg['oslo_messaging_notify_node_names'] = [name]
                globalcfg['memcached_node_ips'] = [name]
            outputs['GlobalConfigExtraMapData'] = globalcfg

        self._create_working_dirs(stack_name.lower())
        output = {'parameter_defaults': outputs}
        with open(endpointmap_file, 'w') as f:
            yaml.safe_dump(output, f, default_flow_style=False)
        return output

    def get_parser(self, prog_name):
        parser = argparse.ArgumentParser(
            description=self.get_description(),
            prog=prog_name,
            add_help=False
        )
        parser.add_argument(
            '--templates', nargs='?', const=constants.TRIPLEO_HEAT_TEMPLATES,
            help=_("The directory containing the Heat templates to deploy"),
            default=constants.TRIPLEO_HEAT_TEMPLATES
        )
        parser.add_argument('--standalone', default=False, action='store_true',
                            help=_("DEPRECATED. The --standalone argument is "
                                   "now deprecated. Standalone deployments "
                                   "can now be run without passing "
                                   "--standalone. "))
        parser.add_argument('--upgrade', default=False, action='store_true',
                            help=_("Upgrade an existing deployment."))
        parser.add_argument('-y', '--yes', default=False, action='store_true',
                            help=_("Skip yes/no prompt (assume yes)."))
        parser.add_argument('--stack',
                            help=_("Name for the ephemeral (one-time create "
                                   "and forget) heat stack."),
                            default='standalone')
        parser.add_argument('--output-dir',
                            dest='output_dir',
                            help=_("Directory to output state, processed heat "
                                   "templates, ansible deployment files."),
                            default=constants.UNDERCLOUD_OUTPUT_DIR)
        parser.add_argument('--output-only',
                            dest='output_only',
                            action='store_true',
                            default=False,
                            help=_("Do not execute the Ansible playbooks. By"
                                   " default the playbooks are saved to the"
                                   " output-dir and then executed.")),
        parser.add_argument('--standalone-role', default='Standalone',
                            help=_("The role to use for standalone "
                                   "configuration when populating the "
                                   "deployment actions."))
        parser.add_argument('-t', '--timeout', metavar='<TIMEOUT>',
                            type=int, default=30,
                            help=_('Deployment timeout in minutes.'))
        parser.add_argument(
            '-e', '--environment-file', metavar='<HEAT ENVIRONMENT FILE>',
            action='append', dest='environment_files',
            help=_('Environment files to be passed to the heat stack-create '
                   'or heat stack-update command. (Can be specified more than '
                   'once.)')
        )
        parser.add_argument(
            '--roles-file', '-r', dest='roles_file',
            help=_(
                'Roles file, overrides the default %s in the t-h-t templates '
                'directory used for deployment. May be an '
                'absolute path or the path relative to the templates dir.'
                ) % constants.UNDERCLOUD_ROLES_FILE
        )
        parser.add_argument(
            '--networks-file', '-n', dest='networks_file',
            help=_(
                'Roles file, overrides the default %s in the t-h-t templates '
                'directory used for deployment. May be an '
                'absolute path or the path relative to the templates dir.'
                ) % constants.STANDALONE_NETWORKS_FILE
        )
        parser.add_argument(
            '--plan-environment-file', '-p',
            help=_('Plan Environment file, overrides the default %s in the '
                   '--templates directory') % constants.PLAN_ENVIRONMENT
        )
        parser.add_argument(
            '--heat-api-port', metavar='<HEAT_API_PORT>',
            dest='heat_api_port',
            default='8006',
            help=_('Heat API port to use for the installers private'
                   ' Heat API instance. Optional. Default: 8006.)')
        )
        parser.add_argument(
            '--heat-user', metavar='<HEAT_USER>',
            dest='heat_user',
            default='heat',
            help=_('User to execute the non-privileged heat-all process. '
                   'Defaults to heat.')
        )
        # TODO(cjeanner) drop that once using oslo.privsep
        parser.add_argument(
            '--deployment-user',
            dest='deployment_user',
            default=os.environ.get('SUDO_USER', 'stack'),
            help=_('User who executes the tripleo deploy command. '
                   'Defaults to $SUDO_USER. If $SUDO_USER is unset '
                   'it defaults to stack.')
        )
        parser.add_argument('--deployment-python-interpreter', default=None,
                            help=_('The path to python interpreter to use for '
                                   'the deployment actions. If not specified '
                                   'the python version of the openstackclient '
                                   'will be used. This may need to be used '
                                   'if deploying on a python2 host from a '
                                   'python3 system or vice versa.'))
        parser.add_argument(
            '--heat-container-image', metavar='<HEAT_CONTAINER_IMAGE>',
            dest='heat_container_image',
            default=constants.DEFAULT_HEAT_CONTAINER,
            help=_('The container image to use when launching the heat-all '
                   'process. Defaults to: {}'.format(
                       constants.DEFAULT_HEAT_CONTAINER))
        )
        parser.add_argument(
            '--heat-native',
            dest='heat_native',
            nargs='?',
            default=None,
            const="true",
            help=_('Execute the heat-all process natively on this host. '
                   'This option requires that the heat-all binaries '
                   'be installed locally on this machine. '
                   'This option is enabled by default which means heat-all is '
                   'executed on the host OS directly.')
        )
        parser.add_argument(
            '--local-ip', metavar='<LOCAL_IP>',
            dest='local_ip',
            help=_('Local IP/CIDR for undercloud traffic. Required.')
        )
        parser.add_argument(
            '--control-virtual-ip', metavar='<CONTROL_VIRTUAL_IP>',
            dest='control_virtual_ip',
            help=_('Control plane VIP. This allows the undercloud installer '
                   'to configure a custom VIP on the control plane.')
        )
        parser.add_argument(
            '--public-virtual-ip', metavar='<PUBLIC_VIRTUAL_IP>',
            dest='public_virtual_ip',
            help=_('Public nw VIP. This allows the undercloud installer '
                   'to configure a custom VIP on the public (external) NW.')
        )
        parser.add_argument(
            '--local-domain', metavar='<LOCAL_DOMAIN>',
            dest='local_domain',
            default='localdomain',
            help=_('Local domain for standalone cloud and its API endpoints')
        )
        parser.add_argument(
            '--cleanup',
            action='store_true', default=False,
            help=_('Cleanup temporary files. Using this flag will '
                   'remove the temporary files used during deployment in '
                   'after the command is run.'),

        )
        parser.add_argument(
            '--hieradata-override', nargs='?',
            help=_('Path to hieradata override file. When it points to a heat '
                   'env file, it is passed in t-h-t via --environment-file. '
                   'When the file contains legacy instack data, '
                   'it is wrapped with <role>ExtraConfig and also '
                   'passed in for t-h-t as a temp file created in '
                   '--output-dir. Note, instack hiera data may be '
                   'not t-h-t compatible and will highly likely require a '
                   'manual revision.')
        )
        parser.add_argument(
            '--keep-running',
            action='store_true',
            default=False,
            help=_('Keep the ephemeral Heat running after the stack operation '
                   'is complete. This is for debugging purposes only. '
                   'The ephemeral Heat can be used by openstackclient with:\n'
                   'OS_AUTH_TYPE=none '
                   'OS_ENDPOINT=http://127.0.0.1:8006/v1/admin '
                   'openstack stack list\n '
                   'where 8006 is the port specified by --heat-api-port.')
        )
        parser.add_argument(
            '--inflight-validations',
            action='store_true',
            default=False,
            dest='inflight',
            help=_('Activate in-flight validations during the deploy. '
                   'In-flight validations provide a robust way to ensure '
                   'deployed services are running right after their '
                   'activation. Defaults to False.')
        )
        parser.add_argument(
            '--transport',
            action='store',
            default='local',
            help=_('Transport mechanism to use for ansible.'
                   'Use "ssh" for multinode deployments. '
                   'Use "local" for standalone deployments. '
                   'Defaults to "local".')
        )
        parser.add_argument(
            '--ansible-forks',
            action='store',
            default=None,
            type=int,
            help=_('The number of Ansible forks to use for the'
                   ' config-download ansible-playbook command.')
        )
        parser.add_argument(
            '--disable-container-prepare',
            action='store_true',
            default=False,
            help=_('Disable the container preparation actions to prevent '
                   'container tags from being updated and new containers '
                   'from being fetched. If you skip this but do not have '
                   'the container parameters configured, the deployment '
                   'action may fail.')
        )

        stack_action_group = parser.add_mutually_exclusive_group()

        stack_action_group.add_argument(
            '--force-stack-update',
            dest='force_stack_update',
            action='store_true',
            default=False,
            help=_("Do a virtual update of the ephemeral "
                   "heat stack (it cannot take real updates). "
                   "New or failed deployments "
                   "always have the stack_action=CREATE. This "
                   "option enforces stack_action=UPDATE."),
        )
        stack_action_group.add_argument(
            '--force-stack-create',
            dest='force_stack_create',
            action='store_true',
            default=False,
            help=_("Do a virtual create of the ephemeral "
                   "heat stack. New or failed deployments "
                   "always have the stack_action=CREATE. This "
                   "option enforces stack_action=CREATE."),
        )
        return parser

    def _process_hieradata_overrides(self, override_file=None,
                                     tripleo_role_name='Standalone',
                                     stack_name='undercloud'):
        """Count in hiera data overrides including legacy formats

        Return a file name that points to processed hiera data overrides file
        """
        if not override_file or not os.path.exists(override_file):
            # we should never get here because there's a check in
            # undercloud_conf but stranger things have happened.
            msg = (_('hieradata_override file could not be found %s') %
                   override_file)
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)

        target = override_file
        with open(target, 'rb') as fb:
            data = fb.read()
        if not data.strip():
            # since an empty file isn't valid yaml, let's be more specific
            msg = (_("hieradata override file (%s) cannot be empty") % target)
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)

        hiera_data = yaml.safe_load(data)
        if not hiera_data:
            msg = (_('Unsupported data format in hieradata override %s') %
                   target)
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)
        self._create_working_dirs(stack_name)

        # NOTE(bogdando): In t-h-t, hiera data should come in wrapped as
        # {parameter_defaults: {UndercloudExtraConfig: ... }}
        extra_config_var = '%sExtraConfig' % tripleo_role_name
        if (extra_config_var not in hiera_data.get('parameter_defaults', {})):
            hiera_override_file = os.path.join(
                self.tht_render, 'tripleo-hieradata-override.yaml')
            self.log.info('Converting hiera overrides for t-h-t from '
                          'legacy format into a file %s' %
                          hiera_override_file)
            with open(hiera_override_file, 'w') as override:
                yaml.safe_dump(
                    {'parameter_defaults': {
                     extra_config_var: hiera_data}},
                    override,
                    default_flow_style=False)
            target = hiera_override_file
        return target

    def _dump_ansible_errors(self, f, name):
        if not os.path.isfile(f):
            return

        failures = None
        with open(f, 'r') as ff:
            try:
                failures = json.load(ff)
            except Exception:
                self.log.error(
                    _('Could not read ansible errors from file %s') % f)

        if not failures or not failures.get(name, {}):
            return

        self.log.error(_('** Found ansible errors for %s deployment! **') %
                       name)
        self.log.error(json.dumps(failures.get(name, {}), indent=1))

    def _standalone_deploy(self, parsed_args):
        extra_env_var = dict()
        if self._is_undercloud_deploy(parsed_args):
            extra_env_var['ANSIBLE_LOG_PATH'] = os.path.join(
                    parsed_args.output_dir, constants.UNDERCLOUD_LOG_FILE)
        if not parsed_args.local_ip:
            msg = _('Please set --local-ip to the correct '
                    'ipaddress/cidr for this machine.')
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)

        if not os.environ.get('HEAT_API_PORT'):
            os.environ['HEAT_API_PORT'] = parsed_args.heat_api_port

        # The main thread runs as root and we drop privs for forked
        # processes below. Only the heat deploy/os-collect-config forked
        # process runs as root.
        if os.geteuid() != 0:
            msg = _("Please run as root.")
            self.log.error(msg)
            raise exceptions.DeploymentError(msg)

        self._run_preflight_checks(parsed_args)

        # prepare working spaces
        self.output_dir = os.path.abspath(parsed_args.output_dir)
        self._create_working_dirs(parsed_args.stack.lower())
        # The state that needs to be persisted between serial deployments
        # and cannot be contained in ephemeral heat stacks or working dirs
        self._create_persistent_dirs()

        # configure puppet
        self._configure_puppet()

        # copy the templates dir in place
        self._populate_templates_dir(parsed_args.templates,
                                     parsed_args.stack.lower())

        # Set default plan if not specified by user
        self._set_default_plan()

        is_complete = False
        try:
            # NOTE(bogdando): Look for the unique virtual update mark matching
            # the heat stack name we are going to create below. If found the
            # mark, consider the stack action is UPDATE instead of CREATE.
            mark_uuid = '_'.join(['update_mark', parsed_args.stack])
            self.stack_update_mark = os.path.join(
                constants.STANDALONE_EPHEMERAL_STACK_VSTATE,
                mark_uuid)

            self._set_stack_action(parsed_args)

            # Launch heat.
            orchestration_client = self._launch_heat(parsed_args)
            # Wait for heat to be ready.
            utils.wait_api_port_ready(parsed_args.heat_api_port)
            # Deploy TripleO Heat templates.
            stack_id = \
                self._deploy_tripleo_heat_templates(orchestration_client,
                                                    parsed_args)

            # Wait for complete..
            status = utils.wait_for_stack_ready(orchestration_client, stack_id,
                                                nested_depth=6)
            if not status:
                message = _("Stack create failed")
                self.log.error(message)
                raise exceptions.DeploymentError(message)

            # download the ansible playbooks and execute them.
            depl_python = utils.get_deployment_python_interpreter(parsed_args)
            self.ansible_dir = \
                self._download_ansible_playbooks(orchestration_client,
                                                 parsed_args.stack,
                                                 parsed_args.standalone_role,
                                                 depl_python)

            # output an file with EndpointMapOverride for use with other stacks
            self._download_stack_outputs(orchestration_client,
                                         parsed_args.stack)

            # Do not override user's custom ansible configuraition file,
            # it may have been pre-created with the tripleo CLI, or the like
            ansible_config = os.path.join(self.output_dir, 'ansible.cfg')
            if not os.path.isfile(ansible_config):
                self.log.warning(
                    _('Generating default ansible config file %s') %
                    ansible_config)
                # FIXME(bogdando): unhardcode key for future
                # multi-node
                ansible.write_default_ansible_cfg(
                    self.ansible_dir,
                    parsed_args.deployment_user,
                    ssh_private_key=None,
                    transport=parsed_args.transport)
            else:
                self.log.warning(
                    _('Using the existing %s for deployment') % ansible_config)
                shutil.copy(ansible_config, self.ansible_dir)

            extra_args = dict()
            if not parsed_args.inflight:
                extra_args = {'skip_tags': 'opendev-validation'}
            # Kill heat, we're done with it now.
            if not parsed_args.keep_running:
                self._kill_heat(parsed_args)
            if not parsed_args.output_only:
                operations = list()
                if parsed_args.upgrade:
                    # Run Upgrade tasks before the deployment
                    operations.append(
                        constants.DEPLOY_ANSIBLE_ACTIONS['upgrade']
                    )
                operations.append(
                    constants.DEPLOY_ANSIBLE_ACTIONS['deploy']
                )
                if parsed_args.upgrade:
                    # Run Post Upgrade tasks after the deployment
                    operations.append(
                        constants.DEPLOY_ANSIBLE_ACTIONS['post-upgrade']
                    )
                    # Run Online Upgrade tasks after the deployment
                    operations.append(
                        constants.DEPLOY_ANSIBLE_ACTIONS['online-upgrade']
                    )
                with utils.Pushd(self.ansible_dir):
                    for operation in operations:
                        for k, v in extra_args.items():
                            if k in operation:
                                operation[k] = ','.join([operation[k], v])
                            else:
                                operation[k] = v
                        utils.run_ansible_playbook(
                            inventory=os.path.join(
                                self.ansible_dir,
                                'inventory.yaml'
                            ),
                            workdir=self.ansible_dir,
                            verbosity=utils.playbook_verbosity(self=self),
                            extra_env_variables=extra_env_var,
                            forks=parsed_args.ansible_forks,
                            **operation)
            is_complete = True
        finally:
            if not parsed_args.keep_running:
                self._kill_heat(parsed_args)
            tar_filename = \
                self._create_install_artifact(parsed_args.deployment_user,
                                              parsed_args.stack.lower())
            if self.ansible_dir:
                self._dump_ansible_errors(
                    os.path.join(self.ansible_dir,
                                 tc_constants.ANSIBLE_ERRORS_FILE),
                    parsed_args.stack)
            self._cleanup_working_dirs(
                cleanup=parsed_args.cleanup,
                user=parsed_args.deployment_user
                )
            if tar_filename:
                self.log.warning('Install artifact is located at %s' %
                                 tar_filename)
            if not is_complete:
                # We only get here on error.
                # Alter the stack virtual state for failed deployments
                if (self.stack_update_mark and
                   not parsed_args.force_stack_update and
                   os.path.isfile(self.stack_update_mark)):
                    self.log.warning(
                        _('The heat stack %s virtual state/action is '
                          'reset to CREATE. Use "--force-stack-update" to '
                          'set it forcefully to UPDATE') % parsed_args.stack)
                    self.log.warning(
                        _('Removing the stack virtual update mark file %s') %
                        self.stack_update_mark)
                    os.remove(self.stack_update_mark)

                self.log.error(DEPLOY_FAILURE_MESSAGE.format(
                    self.heat_launch.install_tmp
                    ))
            else:
                # We only get here if no errors
                if parsed_args.output_only:
                    success_messaging = OUTPUT_ONLY_COMPLETION_MESSAGE
                else:
                    success_messaging = DEPLOY_COMPLETION_MESSAGE

                if not self._is_undercloud_deploy(parsed_args):
                    success_messaging = success_messaging + \
                        STANDALONE_COMPLETION_MESSAGE.format(
                            '~/.config/openstack/clouds.yaml')

                self.log.warning(success_messaging)
                if (self.stack_update_mark and
                   (not parsed_args.output_only or
                       parsed_args.force_stack_update)):
                    # Persist the unique mark file for this stack
                    # Do not update its atime file system attribute to keep its
                    # genuine timestamp for the 1st time the stack state had
                    # been (virtually) changed to match stack_action UPDATE
                    self.log.warning(
                        _('Writing the stack virtual update mark file %s') %
                        self.stack_update_mark)
                    open(self.stack_update_mark, 'w').close()
                elif parsed_args.output_only:
                    self.log.warning(
                        _('Not creating the stack %s virtual update mark file '
                          'in the --output-only mode! Re-run with '
                          '--force-stack-update, if you want to enforce it.') %
                        parsed_args.stack)
                else:
                    self.log.warning(
                        _('Not creating the stack %s virtual update mark '
                          'file') % parsed_args.stack)

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)
        if parsed_args.deployment_user == 'root':
            self.log.warning(
                _("[WARNING] Deployment user is set to 'root'. This may cause "
                  "some deployment files to be located in /root. Please use "
                  "--deployment-user to specify the user you are deploying "
                  "with."))
        try:
            self._standalone_deploy(parsed_args)
        except Exception as ex:
            self.log.error("Exception: %s" % six.text_type(ex))
            self.log.error(traceback.print_exc())
            raise exceptions.DeploymentError(six.text_type(ex))
        finally:
            # Copy clouds.yaml from /etc/openstack so credentials can be
            # read by the deployment user and not only root.
            utils.copy_clouds_yaml(parsed_args.deployment_user)

            # send erase sequence to reset the cmdline if ansible
            # mangled some escape sequences
            utils.reset_cmdline()



#   Copyright 2018 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import fixtures


class DeploymentWorkflowFixture(fixtures.Fixture):

    def _setUp(self):
        super(DeploymentWorkflowFixture, self)._setUp()
        self.mock_get_hosts_and_enable_ssh_admin = self.useFixture(
            fixtures.MockPatch('tripleoclient.workflows.deployment.'
                               'get_hosts_and_enable_ssh_admin')
        ).mock
        self.mock_config_download = self.useFixture(fixtures.MockPatch(
            'tripleoclient.workflows.deployment.config_download')
        ).mock
        self.mock_set_deployment_status = self.useFixture(fixtures.MockPatch(
            'tripleoclient.workflows.deployment.set_deployment_status')
        ).mock


class PlanManagementFixture(fixtures.Fixture):

    def _setUp(self):
        super(PlanManagementFixture, self)._setUp()


class UtilsOvercloudFixture(fixtures.Fixture):

    def _setUp(self):
        super(UtilsOvercloudFixture, self)._setUp()
        self.mock_deploy_tht = self.useFixture(fixtures.MockPatch(
            'tripleoclient.utils.create_tempest_deployer_input')
        ).mock
        self.mock_utils_endpoint = self.useFixture(fixtures.MockPatch(
            'tripleoclient.utils.get_overcloud_endpoint')
        ).mock
        self.mock_update_deployment_status = self.useFixture(
            fixtures.MockPatch(
                'tripleoclient.utils.update_deployment_status')
        ).mock


class UtilsFixture(fixtures.Fixture):

    def _setUp(self):
        super(UtilsFixture, self)._setUp()
        self.wait_for_stack_ready_mock = self.useFixture(fixtures.MockPatch(
            'tripleoclient.utils.wait_for_stack_ready',
            return_value=True)
        ).mock
        self.mock_remove_known_hosts = self.useFixture(fixtures.MockPatch(
            'tripleoclient.utils.remove_known_hosts')
        ).mock
        self.mock_run_ansible_playbook = self.useFixture(fixtures.MockPatch(
            'tripleoclient.utils.run_ansible_playbook')
        ).mock



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import mock

from osc_lib import exceptions
from osc_lib.tests import utils

from tripleoclient.v1 import overcloud_plan_roles


class BaseTestCommand(utils.TestCommand):
    def setUp(self):
        super(BaseTestCommand, self).setUp()

        tc = self.app.client_manager.tripleoclient = mock.Mock()
        tc.object_store.get_object.return_value = (
            {},
            '{"result": [{"name":"Controller","description":"Test desc",'
            '"random": "abcd"},{"name":"Test"}]}'
        )
        tc.object_store.get_container.return_value = (
            'container',
            [
                {
                    "name": "Controller",
                    "description": "Test desc",
                    "random": "abcd",
                    "efg": "123",
                    "ServicesDefault": [
                        "b",
                        "c",
                        "a"
                    ]
                }
            ]
        )
        self.tripleoclient = tc


class TestOvercloudListCurrentRoles(BaseTestCommand):

    def setUp(self):
        super(TestOvercloudListCurrentRoles, self).setUp()

        self.cmd = overcloud_plan_roles.ListRoles(self.app, None)

    @mock.patch(
        'tripleo_common.actions.plan.ListRolesAction.run',
        autospec=True,
        return_value=[]
    )
    def test_list_empty_on_non_default_plan(self, mock_list):
        arglist = ['--name', 'overcast', '--current']
        verifylist = [('name', 'overcast'), ('current', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertEqual(0, len(result[1]))

    @mock.patch(
        'tripleo_common.actions.plan.ListRolesAction.run',
        autospec=True,
        return_value=["ObjectStorage", "Controller"]
    )
    def test_list(self, mock_list):
        arglist = ['--current']
        verifylist = [('name', 'overcloud'), ('current', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertEqual(2, len(result[1]))
        self.assertEqual([('Controller',), ('ObjectStorage',)], result[1])

    @mock.patch(
        'tripleo_common.actions.plan.ListRolesAction.run',
        autospec=True,
        return_value=[
            {
                "name": "Controller",
                "description": "Test desc",
                "random": "abcd"
            },
            {"name": "Test"}
        ]
    )
    def test_list_with_details(self, mock_list):
        parsed_args = self.check_parser(self.cmd,
                                        ['--current', '--detail'],
                                        [])
        result = self.cmd.take_action(parsed_args)

        data = result[1]
        self.assertEqual(2, len(data))

        self.assertEqual(data[0][0], "Controller")
        self.assertEqual(data[0][3], "random: abcd")
        self.assertEqual(data[1][0], "Test")
        self.assertEqual(data[1][3], "")

    @mock.patch(
        'tripleo_common.actions.plan.ListRolesAction.run',
        autospec=True,
        return_value=[]
    )
    def test_list_with_details_empty(self, mock_list):
        parsed_args = self.check_parser(self.cmd,
                                        ['--current', '--detail'],
                                        [])
        result = self.cmd.take_action(parsed_args)

        self.assertEqual(0, len(result[1]))

    @mock.patch(
        'tripleo_common.actions.plan.ListRolesAction.run',
        autospec=True,
        return_value=[
            {"name": "Compute"},
            {"name": "Random"},
            {"name": "BlockStorage", "ServicesDefault": ["c", "b", "a"]}
        ]
    )
    def test_list_with_details_sorted(self, mock_list):

        parsed_args = self.check_parser(self.cmd,
                                        ['--current', '--detail'],
                                        [])
        result = self.cmd.take_action(parsed_args)

        self.assertEqual(3, len(result[1]))

        # Test main list sorted
        self.assertEqual(result[1][0][0], "BlockStorage")
        self.assertEqual(result[1][1][0], "Compute")
        self.assertEqual(result[1][2][0], "Random")

        # Test service sublist sorted
        self.assertEqual(result[1][0][2], "a\nb\nc")


class TestOvercloudShowRole(BaseTestCommand):

    def setUp(self):
        super(TestOvercloudShowRole, self).setUp()

        self.cmd = overcloud_plan_roles.ShowRole(self.app, None)

        self.app.client_manager.tripleoclient = self.tripleoclient

    @mock.patch(
        'tripleo_common.actions.plan.ListRolesAction.run',
        autospec=True,
        return_value=[]
    )
    def test_role_not_found(self, mock_list):
        arglist = ['--name', 'overcast', 'doesntexist']
        verifylist = [('name', 'overcast'), ('role', 'doesntexist')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)
